# pylint: disable=protected-access,unused-argument

import os
import glob

import radical.utils as ru

from ..                 import states as s
from ..task_description import RAPTOR_MASTER, RAPTOR_WORKER, TASK_EXECUTABLE
from .session           import fetch_json

_debug      = os.environ.get('RP_PROF_DEBUG')
_node_index = dict()


# ------------------------------------------------------------------------------
#
# pilot and task activities: core hours are derived by multiplying the
# respective time durations with pilot size / task size.  The 'idle'
# utilization and the 'agent' utilization are derived separately.
#
# Note that durations should add up to the `x_total` generations to ensure
# accounting for the complete task/pilot utilization.
#
# An updated list of events is available at docs/source/events.md


PILOT_DURATIONS = {
    'provide' : {
        'total'     : [{ru.EVENT: 'bootstrap_0_start'},
                       {ru.EVENT: 'bootstrap_0_stop' }]
    },
    # times between PMGR_ACTIVE and the termination command are not
    # considered pilot specific consumptions.  If some resources remain
    # unused during that time, it is either due to inefficiencies of
    # workload management (accounted for in the task consumption metrics),
    # or the pilot is starving for workload.
    'consume' : {
        'boot'      : [{ru.EVENT: 'bootstrap_0_start'},
                       {ru.EVENT: 'bootstrap_0_ok'   }],
        'setup_1'   : [{ru.EVENT: 'bootstrap_0_ok'   },
                       {ru.STATE: s.PMGR_ACTIVE      }],
        'idle'      : [{ru.STATE: s.PMGR_ACTIVE      },
                       {ru.EVENT: 'cmd'              ,
                        ru.MSG  : 'cancel_pilot'     }],
        'term'      : [{ru.EVENT: 'cmd'              ,
                        ru.MSG  : 'cancel_pilot'     },
                       {ru.EVENT: 'bootstrap_0_stop' }],
        'agent'     : [{ru.EVENT: 'sub_agent_start'  },
                       {ru.EVENT: 'sub_agent_stop'   }],
    },
    # FIXME: separate out DVM startup time
    #   'rte'       : [{ru.STATE: s.PMGR_ACTIVE    },
    #                  {ru.STATE: s.PMGR_ACTIVE    }],
    #   'setup_2'   : [{ru.STATE: s.PMGR_ACTIVE    },
    #                  {ru.STATE: s.PMGR_ACTIVE    }],
    #
    # resources on agent nodes are consumed for all of the pilot's lifetime
    'agent' : {
        'total'     : [{ru.EVENT: 'bootstrap_0_start'},
                       {ru.EVENT: 'bootstrap_0_stop' }]
    }
}


# The set of default task durations that are available for every task
# description, default resource configuration, and default scheduler and
# launcher.
TASK_DURATIONS_DEFAULT = {
    'consume' : {
        'exec_queue'  : [{ru.EVENT: 'schedule_ok'            },
                         {ru.STATE: s.AGENT_EXECUTING        }],
        'exec_prep'   : [{ru.STATE: s.AGENT_EXECUTING        },
                         {ru.EVENT: 'task_run_start'         }],
        'exec_rp'     : [{ru.EVENT: 'task_run_start'         },
                         {ru.EVENT: 'launch_start'           }],
        'exec_sh'     : [{ru.EVENT: 'launch_start'           },
                         {ru.EVENT: 'exec_start'             }],
        'exec_cmd'    : [{ru.EVENT: 'exec_start'             },
                         {ru.EVENT: 'exec_stop'              }],
        'term_sh'     : [{ru.EVENT: 'exec_stop'              },
                         {ru.EVENT: 'launch_stop'            }],
        'term_rp'     : [{ru.EVENT: 'launch_stop'            },
                         {ru.EVENT: 'task_run_stop'          }],
        'unschedule'  : [{ru.EVENT: 'task_run_stop'          },
                         {ru.EVENT: 'unschedule_stop'        }]

      # if we have cmd_start / cmd_stop (exec_sh including all preparations):
      # 'exec_sh'     : [{ru.EVENT: 'task_start'             },
      #                  {ru.EVENT: 'cmd_start'              }],
      # 'exec_cmd'    : [{ru.EVENT: 'cmd_start'              },
      #                  {ru.EVENT: 'cmd_stop'               }],
      # 'term_sh'     : [{ru.EVENT: 'cmd_stop'               },
      #                  {ru.EVENT: 'task_stop'              }],
    }
}


# The set of default task durations augmented with the durations of the app
# events. App events are generated by RADICAL Synapse and by the script
# `radical-pilot-hello.sh`. The latter is useful for testing as
# a sleep command drop-in.
TASK_DURATIONS_APP = {
    'consume' : {
        'exec_queue'  : [{ru.EVENT: 'schedule_ok'            },
                         {ru.STATE: s.AGENT_EXECUTING        }],
        'exec_prep'   : [{ru.STATE: s.AGENT_EXECUTING        },
                         {ru.EVENT: 'task_run_start'         }],
        'exec_rp'     : [{ru.EVENT: 'task_run_start'         },
                         {ru.EVENT: 'launch_start'           }],
        'exec_sh'     : [{ru.EVENT: 'launch_start'           },
                         {ru.EVENT: 'exec_start'             }],
        'init_app'    : [{ru.EVENT: 'exec_start'             },
                         {ru.EVENT: 'rank_start'             }],
        'exec_cmd'    : [{ru.EVENT: 'rank_start'             },
                         {ru.EVENT: 'rank_stop'              }],
        'term_app'    : [{ru.EVENT: 'rank_stop'              },
                         {ru.EVENT: 'exec_stop'              }],
        'term_sh'     : [{ru.EVENT: 'exec_stop'              },
                         {ru.EVENT: 'launch_stop'            }],
        'term_rp'     : [{ru.EVENT: 'launch_stop'            },
                         {ru.EVENT: 'task_run_stop'          }],
        'unschedule'  : [{ru.EVENT: 'task_run_stop'          },
                         {ru.EVENT: 'unschedule_stop'        }]
    }
}


# FIXME: TASK_DURATIONS_PRTE
# The set of default task durations with the durations generated when using
# PRRTE as a launch method.


# ----------------------------------------------------------------------------
#
def _convert_sdurations(sdurations):
    '''
    Converts a collection of durations expressed in short form to the same
    collection of durations expressed in long form.

    Definitions:

    - Short form collection: one dictionary of short form durations
    - Long form: one dictionary of long form durations.

    Args:

        sdurations (dict): a collections of durations in short form

    Return:

        ldurations (dict): a collection of long form durations

    Example:

        sdurations = {'name_of_duration': [{'STATE': s.STATE_NAME},
                                           {'EVENT': 'event_name'}]}
        ldurations = {'name_of_duration': [{ru.EVENT: 'state',
                                            ru.STATE: s.STATE_NAME},
                                           {ru.EVENT: 'event_name',
                                            ru.STATE: None}]}
        sdurations = {'name_of_duration': [{'STATE': s.STATE_NAME},
                                           [{'EVENT': 'event_name'},
                                            {'STATE': s.STATE_NAME}]]}
        ldurations = {'name_of_duration': [{ru.EVENT: 'state',
                                            ru.STATE: s.STATE_NAME},
                                           [{ru.EVENT: 'event_name',
                                             ru.STATE: None},
                                            {ru.EVENT: 'state',
                                             ru.STATE: s.STATE_NAME}]]}
        sdurations = {'name_of_duration': [{'STATE': s.STATE_NAME},
                                           {'MSG': 'message_name'}]}
        ldurations = {'name_of_duration': [{ru.EVENT: 'state',
                                            ru.STATE: s.STATE_NAME},
                                           {ru.EVENT: 'cmd',
                                            ru.MSG: 'message_name'}]}
    '''

    ldurations = dict()

    for k,v in sdurations.items():

        ldurations[k] = list()
        for ts in v:

            if isinstance(ts, dict):
                ldurations[k].append(_expand_sduration(ts))

            if isinstance(ts, list):
                lds = list()
                for i in ts:
                    lds.append(_expand_sduration(i))
                ldurations[k].append(lds)

    return ldurations


# ----------------------------------------------------------------------------
#
def _expand_sduration(sduration):
    '''
    Expands a duration expressed in short form to its long form for the
    timestamp types `ru.STATE`, `ru.EVENT` and `ru.MSG`.

    Definitions:

    - Short form duration: one dictionary containing a state or event name.
    - Long form duration: one dictionary containing two keys, one of type
      `ru.EVENT` and one of type `ru.STATE`. The `ru.EVENT` key has a string
      value while the `ru.STATE` key has a `s.STATE_NAME` object as its value.

    Args:

        sduration (dict): a duration in short form

    Return:

        lduration (dict): sduration in long form

    Example:

        sduration = {'STATE': s.STATE_NAME}
        lduration = {ru.EVENT: 'state', ru.STATE: s.STATE_NAME}
        sduration = {'EVENT': 'event_name'}
        lduration = {ru.EVENT: 'event_name', ru.STATE: None}
        sduration = {'MSG': 'mesage_name'}
        lduration = {ru.EVENT: 'cmd', ru.MSG: 'message_name'}
    '''

    # Allow durations with both ru.EVENT and ru.STATE.
    tt = list(sduration.keys())
    if len(tt) == 1 and tt[0] not in ['STATE', 'EVENT', 'MSG']:
        raise Exception('unknown timestamp type: %s' % tt)
    if len(tt) == 2:
        return sduration
    if len(tt) > 2:
        raise Exception('invalid duration: too many timestamps (%s)' % tt)

    # Expand known timestamps.
    lduration = None

    for k,v in sduration.items():
        if k == 'STATE':
            lduration = {ru.EVENT: 'state', ru.STATE: v}
        elif k == 'EVENT':
            lduration = {ru.EVENT: v, ru.STATE: None}
        elif k == 'MSG':
            lduration = {ru.EVENT: 'cmd', ru.MSG: v}

    return lduration


# Set of default pilot durations for RADICAL-Analytics. All the durations
# are contiguos.
# NOTE: _init durations are most often 0.
PILOT_DURATIONS_DEBUG_SHORT = {
    'p_pmgr_create'           : [{'STATE': s.NEW                   },
                                 {'STATE': s.PMGR_LAUNCHING_PENDING}],
    'p_pmgr_launching_init'   : [{'STATE': s.PMGR_LAUNCHING_PENDING},
                                 {'STATE': s.PMGR_LAUNCHING        }],
    'p_pmgr_launching'        : [{'STATE': s.PMGR_LAUNCHING        },
                                 {'EVENT': 'staging_in_start'      }],
    'p_pmgr_stage_in'         : [{'EVENT': 'staging_in_start'      },
                                 {'EVENT': 'staging_in_stop'       }],
    'p_pmgr_submission_init'  : [{'EVENT': 'staging_in_stop'       },
                                 {'EVENT': 'submission_start'      }],
    'p_pmgr_submission'       : [{'EVENT': 'submission_start'      },
                                 {'EVENT': 'submission_stop'       }],
    'p_pmgr_scheduling_init'  : [{'EVENT': 'submission_stop'       },
                                 {'STATE': s.PMGR_ACTIVE_PENDING   }],
    # batch system queue time
    'p_pmgr_scheduling'       : [{'STATE': s.PMGR_ACTIVE_PENDING   },
                                 {'EVENT': 'bootstrap_0_start'     }],
    'p_agent_ve_setup_init'   : [{'EVENT': 'bootstrap_0_start'     },
                                 {'EVENT': 've_setup_start'        }],
    'p_agent_ve_setup'        : [{'EVENT': 've_setup_start'        },
                                 {'EVENT': 've_setup_stop'         }],
    'p_agent_ve_activate_init': [{'EVENT': 've_setup_stop'         },
                                 {'EVENT': 've_activate_start'     }],
    'p_agent_ve_activate'     : [{'EVENT': 've_activate_start'     },
                                 {'EVENT': 've_activate_stop'      }],
    'p_agent_install_init'    : [{'EVENT': 've_activate_stop'      },
                                 {'EVENT': 'rp_install_start'      }],
    'p_agent_install'         : [{'EVENT': 'rp_install_start'      },
                                 {'EVENT': 'rp_install_stop'       }],
    'p_agent_launching'       : [{'EVENT': 'rp_install_stop'       },
                                 {'STATE': s.PMGR_ACTIVE           }],
    'p_agent_terminate_init'  : [{'STATE': s.PMGR_ACTIVE           },
                                 {'MSG'  : 'cancel_pilot'          }],
    'p_agent_terminate'       : [{'MSG'  : 'cancel_pilot'          },
                                 {'EVENT': 'bootstrap_0_stop'      }],
    # total pilot runtime
    'p_agent_finalize'        : [{'EVENT': 'bootstrap_0_stop'      },
                                 [{'STATE': s.DONE                 },
                                  {'STATE': s.CANCELED             },
                                  {'STATE': s.FAILED               }]],
    'p_agent_runtime'         : [{'EVENT': 'bootstrap_0_start'     },
                                 {'EVENT': 'bootstrap_0_stop'      }]
}

PILOT_DURATIONS_DEBUG = _convert_sdurations(PILOT_DURATIONS_DEBUG_SHORT)


# Debug pilot durations tagged with keys taht can be used when calculating
# resource utilization.
# TODO: add the 'client' tag to relevant resource utilization methods.
_pdd = PILOT_DURATIONS_DEBUG
PILOT_DURATIONS_DEBUG_RU = {
    'provide' : {
        'p_agent_runtime'         : _pdd['p_agent_runtime']
    },
    'client'  : {
        'p_pmgr_create'           : _pdd['p_pmgr_create'],
        'p_pmgr_launching_init'   : _pdd['p_pmgr_launching_init'],
        'p_pmgr_launching'        : _pdd['p_pmgr_launching'],
        'p_pmgr_stage_in'         : _pdd['p_pmgr_stage_in'],
        'p_pmgr_submission_init'  : _pdd['p_pmgr_submission_init'],
        'p_pmgr_submission'       : _pdd['p_pmgr_submission'],
        'p_pmgr_scheduling_init'  : _pdd['p_pmgr_scheduling_init'],
        'p_pmgr_scheduling'       : _pdd['p_pmgr_scheduling'],
        'p_agent_finalize'        : _pdd['p_agent_finalize']
    },
    'consume' : {
        'p_agent_ve_setup_init'   : _pdd['p_agent_ve_setup_init'],
        'p_agent_ve_setup'        : _pdd['p_agent_ve_setup'],
        'p_agent_ve_activate_init': _pdd['p_agent_ve_activate_init'],
        'p_agent_ve_activate'     : _pdd['p_agent_ve_activate'],
        'p_agent_install_init'    : _pdd['p_agent_install_init'],
        'p_agent_install'         : _pdd['p_agent_install'],
        'p_agent_launching'       : _pdd['p_agent_launching'],
        'p_agent_terminate_init'  : _pdd['p_agent_terminate_init'],
        'p_agent_terminate'       : _pdd['p_agent_terminate']
    },
    'agent'   : {
        'p_agent_runtime'         : _pdd['p_agent_runtime']
    }
}


# Set of default task durations for RADICAL-Analytics. All the durations
# are contiguous.
TASK_DURATIONS_DEBUG_SHORT = {
    't_tmgr_create'              : [{'STATE': s.NEW                         },
                                    {'STATE': s.TMGR_SCHEDULING_PENDING     }],
    't_tmgr_schedule_queue'      : [{'STATE': s.TMGR_SCHEDULING_PENDING     },
                                    {'STATE': s.TMGR_SCHEDULING             }],
    't_tmgr_schedule'            : [{'STATE': s.TMGR_SCHEDULING             },
                                    {'STATE': s.TMGR_STAGING_INPUT_PENDING  }],
    # push to mongodb
    't_tmgr_stage_in_queue'      : [{'STATE': s.TMGR_STAGING_INPUT_PENDING  },
                                    {'STATE': s.TMGR_STAGING_INPUT          }],
    # wait in mongodb
    't_tmgr_stage_in'            : [{'STATE': s.TMGR_STAGING_INPUT          },
                                    {'STATE': s.AGENT_STAGING_INPUT_PENDING }],
    # pull from mongodb
    't_agent_stage_in_queue'     : [{'STATE': s.AGENT_STAGING_INPUT_PENDING },
                                    {'STATE': s.AGENT_STAGING_INPUT         }],
    't_agent_stage_in'           : [{'STATE': s.AGENT_STAGING_INPUT         },
                                    {'STATE': s.AGENT_SCHEDULING_PENDING    }],
    't_agent_schedule_queue'     : [{'STATE': s.AGENT_SCHEDULING_PENDING    },
                                    {'STATE': s.AGENT_SCHEDULING            }],
    't_agent_schedule'           : [{'STATE': s.AGENT_SCHEDULING            },
                                    {'STATE': s.AGENT_EXECUTING_PENDING     }],
    't_agent_execute_queue'      : [{'STATE': s.AGENT_EXECUTING_PENDING     },
                                    {'STATE': s.AGENT_EXECUTING             }],
    't_agent_execute_prepare'    : [{'STATE': s.AGENT_EXECUTING             },
                                    {'EVENT': 'task_mkdir'                  }],
    't_agent_execute_mkdir'      : [{'EVENT': 'task_mkdir'                  },
                                    {'EVENT': 'task_mkdir_done'             }],
    't_agent_execute_layer_start': [{'EVENT': 'task_mkdir_done'             },
                                    {'EVENT': 'task_run_start'              }],
    # ssh, mpi, ...
    't_agent_execute_layer'      : [{'EVENT': 'task_run_start'              },
                                    {'EVENT': 'task_run_ok'                 }],
    # PROBLEM: discontinuity
    't_agent_lm_start'           : [{'EVENT': 'task_run_start'              },
                                    {'EVENT': 'launch_start'                }],
    't_agent_lm_submit'          : [{'EVENT': 'launch_start'                },
                                    {'EVENT': 'exec_start'                  }],
    't_agent_lm_execute'         : [{'EVENT': 'exec_start'                  },
                                    {'EVENT': 'exec_stop'                   }],
    't_agent_lm_stop'            : [{'EVENT': 'exec_stop'                   },
                                    {'EVENT': 'task_run_stop'               }],
    't_agent_stage_out_start'    : [{'EVENT': 'task_run_stop'               },
                                    {'STATE': s.AGENT_STAGING_OUTPUT_PENDING}],
    't_agent_stage_out_queue'    : [{'STATE': s.AGENT_STAGING_OUTPUT_PENDING},
                                    {'STATE': s.AGENT_STAGING_OUTPUT        }],
    't_agent_stage_out'          : [{'STATE': s.AGENT_STAGING_OUTPUT        },
                                    {'STATE': s.TMGR_STAGING_OUTPUT_PENDING }],
    # push/pull mongodb
    't_agent_push_to_tmgr'       : [{'STATE': s.TMGR_STAGING_OUTPUT_PENDING },
                                    {'STATE': s.TMGR_STAGING_OUTPUT         }],
    't_tmgr_destroy'             : [{'STATE': s.TMGR_STAGING_OUTPUT         },
                                    [{'STATE': s.DONE                       },
                                     {'STATE': s.CANCELED                   },
                                     {'STATE': s.FAILED                     }]],
    't_agent_unschedule'         : [{'EVENT': 'unschedule_start'            },
                                    {'EVENT': 'unschedule_stop'             }]
}

TASK_DURATIONS_DEBUG = _convert_sdurations(TASK_DURATIONS_DEBUG_SHORT)


# Debug task durations tagged with keys taht can be used when calculating
# resource utilization.
# TODO: add the 'client' tag to relevant resource utilization methods.
_udd = TASK_DURATIONS_DEBUG
TASK_DURATIONS_DEBUG_RU = {
    'client' : {
        't_tmgr_create'              : _udd['t_tmgr_create'],
        't_tmgr_schedule_queue'      : _udd['t_tmgr_schedule_queue'],
        't_tmgr_schedule'            : _udd['t_tmgr_schedule'],
        't_tmgr_stage_in_queue'      : _udd['t_tmgr_stage_in_queue'],
        't_tmgr_stage_in'            : _udd['t_tmgr_stage_in'],
        't_tmgr_destroy'             : _udd['t_tmgr_destroy'],
        't_agent_unschedule'         : _udd['t_agent_unschedule']
    },
    'consume'  : {
        't_agent_stage_in_queue'     : _udd['t_agent_stage_in_queue'],
        't_agent_stage_in'           : _udd['t_agent_stage_in'],
        't_agent_schedule_queue'     : _udd['t_agent_schedule_queue'],
        't_agent_schedule'           : _udd['t_agent_schedule'],
        't_agent_execute_queue'      : _udd['t_agent_execute_queue'],
        't_agent_execute_prepare'    : _udd['t_agent_execute_prepare'],
        't_agent_execute_mkdir'      : _udd['t_agent_execute_mkdir'],
        't_agent_execute_layer_start': _udd['t_agent_execute_layer_start'],
        't_agent_execute_layer'      : _udd['t_agent_execute_layer'],
        't_agent_lm_start'           : _udd['t_agent_lm_start'],
        't_agent_lm_submit'          : _udd['t_agent_lm_submit'],
        't_agent_lm_execute'         : _udd['t_agent_lm_execute'],
        't_agent_lm_stop'            : _udd['t_agent_lm_stop'],
        't_agent_stage_out_start'    : _udd['t_agent_stage_out_start'],
        't_agent_stage_out_queue'    : _udd['t_agent_stage_out_queue'],
        't_agent_stage_out'          : _udd['t_agent_stage_out'],
        't_agent_push_to_tmgr'       : _udd['t_agent_push_to_tmgr'],
    }
}


# ------------------------------------------------------------------------------
#
def get_hostmap(profile):
    '''
    We abuse the profile combination to also derive a pilot-host map, which
    will tell us on what exact host each pilot has been running.  To do so, we
    check for the PMGR_ACTIVE advance event in agent.0.prof, and use the NTP
    sync info to associate a hostname.
    '''
    # FIXME: This should be replaced by proper hostname logging
    #        in `pilot.resource_details`.

    hostmap = dict()  # map pilot IDs to host names
    for entry in profile:
        if  entry[ru.EVENT] == 'hostname':
            hostmap[entry[ru.UID]] = entry[ru.MSG]

    return hostmap


# ------------------------------------------------------------------------------
#
def get_hostmap_deprecated(profiles):
    '''
    This method mangles combine_profiles and get_hostmap, and is deprecated.
    At this point it only returns the hostmap
    '''

    hostmap = dict()  # map pilot IDs to host names
    for pname, prof in profiles.items():

        if not len(prof):
            continue

        if not prof[0][ru.MSG]:
            continue

        host, ip, _, _, _ = prof[0][ru.MSG].split(':')
        host_id = '%s:%s' % (host, ip)

        for row in prof:

            if 'agent.0.prof' in pname    and \
                row[ru.EVENT] == 'advance' and \
                row[ru.STATE] == s.PMGR_ACTIVE:
                hostmap[row[ru.UID]] = host_id
                break

    return hostmap


# ------------------------------------------------------------------------------
#
def get_session_profile(sid, src=None):

    if not src:
        src = '%s/%s' % (os.getcwd(), sid)

    if os.path.exists(src):
        # we have profiles locally
        profiles  = glob.glob('%s/*.prof'    % src)
        profiles += glob.glob('%s/**/*.prof' % src)
    else:
        # need to fetch profiles
        from .session import fetch_profiles
        profiles = fetch_profiles(sid=sid, skip_existing=True)

    #  filter out some frequent, but uninteresting events
    efilter = {ru.EVENT: [
                        # 'get',
                          'publish',
                          'schedule_skip',
                          'schedule_fail',
                          'staging_stderr_start',
                          'staging_stderr_stop',
                          'staging_stdout_start',
                          'staging_stdout_stop',
                          'staging_uprof_start',
                          'staging_uprof_stop',
                          'update_pushed',
                         ]}

    profiles          = ru.read_profiles(profiles, sid, efilter=efilter)
    profile, accuracy = ru.combine_profiles(profiles)
    profile           = ru.clean_profile(profile, sid, s.FINAL, s.CANCELED)
    hostmap           = get_hostmap(profile)

    if not hostmap:
        # FIXME: legacy host notation - deprecated
        hostmap = get_hostmap_deprecated(profiles)

    return profile, accuracy, hostmap


# ------------------------------------------------------------------------------
#
def get_session_description(sid, src=None, dburl=None):
    '''
    This will return a description which is usable for radical.analytics
    evaluation.  It informs about
      - set of stateful entities
      - state models of those entities
      - event models of those entities (maybe)
      - configuration of the application / module

    If `src` is given, it is interpreted as path to search for session
    information (json dump).  `src` defaults to `$PWD/$sid`.

    if `dburl` is given, its value is used to fetch session information from
    a database.  The dburl value defaults to `RADICAL_PILOT_DBURL`.
    '''

    if not src:
        src = '%s/%s' % (os.getcwd(), sid)

    if os.path.isfile('%s/%s.json' % (src, sid)):
        json = ru.read_json('%s/%s.json' % (src, sid))
    else:
        ftmp = fetch_json(sid=sid, dburl=dburl, tgt=src, skip_existing=True)
        json = ru.read_json(ftmp)

    # make sure we have uids
    # FIXME v0.47: deprecate
    def fix_json(json):
        def fix_uids(json):
            if isinstance(json, list):
                for elem in json:
                    fix_uids(elem)
            elif isinstance(json, dict):
                if 'taskmanager' in json and 'tmgr' not in json:
                    json['tmgr'] = json['taskmanager']
                if 'pilotmanager' in json and 'pmgr' not in json:
                    json['pmgr'] = json['pilotmanager']
                if '_id' in json and 'uid' not in json:
                    json['uid'] = json['_id']
                    if 'cfg' not in json:
                        json['cfg'] = dict()
                for v in json.values():
                    fix_uids(v)
        fix_uids(json)
    fix_json(json)

    assert sid == json['session'][0]['uid'], 'sid inconsistent'

    ret             = dict()
    ret['entities'] = dict()

    tree      = dict()
    tree[sid] = {'uid'      : sid,
                 'etype'    : 'session',
                 'cfg'      : json['session'][0]['cfg'],
                 'has'      : ['tmgr', 'pmgr'],
                 'children' : list()
                }

    for pmgr in sorted(json['pmgr'], key=lambda k: k['uid']):
        uid = pmgr['uid']
        tree[sid]['children'].append(uid)
        tree[uid] = {'uid'      : uid,
                     'etype'    : 'pmgr',
                     'cfg'      : pmgr['cfg'],
                     'has'      : ['pilot'],
                     'children' : list()
                    }

    for tmgr in sorted(json['tmgr'], key=lambda k: k['uid']):
        uid = tmgr['uid']
        tree[sid]['children'].append(uid)
        tree[uid] = {'uid'      : uid,
                     'etype'    : 'tmgr',
                     'cfg'      : tmgr['cfg'],
                     'has'      : ['task'],
                     'children' : list()
                    }
        # also inject the pilot description, and resource specifically
        tree[uid]['description'] = dict()

    for pilot in sorted(json['pilot'], key=lambda k: k['uid']):
        uid  = pilot['uid']
        pmgr = pilot['pmgr']
        pilot['cfg']['resource_details'] = pilot['resource_details']
        tree[pmgr]['children'].append(uid)
        tree[uid] = {'uid'        : uid,
                     'etype'      : 'pilot',
                     'cfg'        : pilot['cfg'],
                     'resources'  : pilot['resources'],
                     'description': pilot['description'],
                     'has'        : ['task'],
                     'children'   : list()
                    }
        # also inject the pilot description, and resource specifically

    for task in sorted(json['task'], key=lambda k: k['uid']):
        uid  = task['uid']
        tmgr = task.get('tmgr')
        if tmgr:
            tree[tmgr]['children'].append(uid)

        pid  = task['pilot']
        tree[pid]['children'].append(uid)

        if 'resources' not in task:
            td = task['description']
            task['resources'] = {'cpu': td['ranks'] * td['cores_per_rank'],
                                 'gpu': td['ranks'] * td['gpus_per_rank']}

        # we determine the entity type by the task mode
        etype = task.get('etype')
        mode  = task['description'].get('mode')
        if not etype:
            mode = task['description']['mode']
            if mode in [RAPTOR_MASTER, RAPTOR_WORKER]:
                etype = mode
            elif mode in [TASK_EXECUTABLE]:
                etype = 'task'
            else:
                etype = 'raptor.task'

        if _debug:
            print('%-30s [%-30s] -> %-30s' % (uid, mode, etype))

        tree[uid] = {'uid'         : uid,
                     'etype'       : etype,
                     'cfg'         : task,
                     'resources'   : task['resources'],
                     'description' : task['description'],
                     'has'         : list(),
                     'children'    : list()
                    }
        # remove duplicate
        del tree[uid]['cfg']['description']

    ret['tree'] = tree

    ret['entities']['pilot']   = {'state_model'  : s._pilot_state_values,
                                  'state_values' : s._pilot_state_inv_full,
                                  'event_model'  : dict()}
    ret['entities']['task']    = {'state_model'  : s._task_state_values,
                                  'state_values' : s._task_state_inv_full,
                                  'event_model'  : dict()}
    ret['entities']['session'] = {'state_model'  : None,  # has no states
                                  'state_values' : None,
                                  'event_model'  : dict()}

    ret['config'] = dict()  # session config goes here

    return ret


# ------------------------------------------------------------------------------
#
def get_node_index(node_list, node_uid, pn):

    if not _node_index:
        for idx,n in enumerate(node_list):
            _node_index[n['node_id']] = idx

    r0 = _node_index[node_uid] * pn
    r1 = r0 + pn - 1

    return [r0, r1]


# ------------------------------------------------------------------------------
#
def get_duration(thing, dur):


    for e in dur:
        if ru.STATE in e and ru.EVENT not in e:
            e[ru.EVENT] = 'state'

    t0 = thing.timestamps(event=dur[0])
    t1 = thing.timestamps(event=dur[1])

    if not len(t0) or not len(t1):
        return [None, None]

    return (t0[0], t1[-1])


# ------------------------------------------------------------------------------
#
def cluster_resources(resources):
    # resources is a list of
    #   - single index (single core of gpu
    #   - [r0, r1] tuples (ranges of core, gpu indexes)
    # cluster continuous stretches of resources

    ret = list()
    idx = set()
    for r in resources:
        if isinstance(r, int):
            idx.add(r)
        else:
            for i in range(r[0], r[1] + 1):
                idx.add(i)

    r0 = None
    r1 = None
    for i in sorted(list(idx)):
        if r0 is None:
            r0 = i
            continue
        if r1 is None:
            if i == r0 + 1:
                r1 = i
                continue
            ret.append([r0, r0])
            r0 = None
            continue
        if i == r1 + 1:
            r1 = i
            continue
        ret.append([r0, r1])
        r0 = i
        r1 = None

    if r0 is not None:
        if r1 is not None:
            ret.append([r0, r1])
        else:
            ret.append([r0, r0])

    return ret


# ------------------------------------------------------------------------------
#
def _get_pilot_provision(pilot, rtype):

    pid = pilot.uid
    ret = dict()
    rnd = 'cores_per_node'

    if rtype == 'gpu':
        rnd = 'gpus_per_node'

    pn = pilot.cfg['resource_details']['rm_info'][rnd]

    nodes, _, _ = _get_nodes(pilot)

    for metric in PILOT_DURATIONS['provide']:

        boxes  = list()
        t0, t1 = get_duration(pilot, PILOT_DURATIONS['provide'][metric])

        if t0 is None:
            t0 = pilot.events [0][ru.TIME]
            t1 = pilot.events[-1][ru.TIME]

        for node in nodes:
            r0, r1 = get_node_index(nodes, node['node_id'], pn)
            boxes.append([t0, t1, r0, r1])

        ret['total'] = {pid: boxes}

    return ret


# ------------------------------------------------------------------------------
#
def get_provided_resources(session, rtype='cpu'):
    '''
    For all ra.pilots, return the amount and time of the type of resources
    provided. This computes sets of 4-tuples of the form: [t0, t1, r0, r1]
    where:

        t0: time, begin of resource provision
        t1: time, begin of resource provision
        r0: int,  index of resources provided (min)
        r1: int,  index of resources provided (max)

    The tuples are formed so that t0 to t1 and r0 to r1 are continuous.
    '''

    if rtype not in ['cpu', 'gpu']:
        raise Exception('unknown resource type: %s' % rtype)

    provided = dict()
    for p in session.get(etype='pilot'):

        data = _get_pilot_provision(p, rtype)

        for metric in data:

            if metric not in provided:
                provided[metric] = dict()

            for uid in data[metric]:
                provided[metric][uid] = data[metric][uid]

    return provided


# ------------------------------------------------------------------------------
#
def get_consumed_resources(session, rtype='cpu', tdurations=None):
    '''
    For all ra.pilot or ra.task entities, return the amount and time of
    resources consumed.  A consumed resource is characterized by:

      - a resource type (we know about cores and gpus)
      - a metric name (what the resource was used for)
      - a list of 4-tuples of the form: [t0, t1, r0, r1]
          - t0: time, begin of resource consumption
          - t1: time, begin of resource consumption
          - r0: int,  index of resources consumed (min)
          - r1: int,  index of resources consumed (max)
        The tuples are formed so that t0 to t1 and r0 to r1 are continuous.

    An entity can consume different resources under different metrics - but the
    returned consumption specs will never overlap, meaning, that any resource is
    accounted for exactly one metric at any point in time.  The returned
    structure has the following overall form:

        {
          'metric_1' : {
              uid_1 : [[t0, t1, r0, r1],
                       [t2, t3, r2, r3],
                       ...
                      ],
              uid_2 : ...
          },
          'metric_2' : ...
        }
    '''

    log = ru.Logger('radical.pilot.utils')

    consumed = dict()
    for e in session.get(etype=['pilot', 'task']):

        if   e.etype == 'pilot': data = _get_pilot_consumption(e, rtype)
        elif e.etype == 'task' : data = _get_task_consumption(session, e, rtype,
                                                              tdurations)

        for metric in data:

            if metric not in consumed:
                consumed[metric] = dict()

            for uid in data[metric]:
                consumed[metric][uid] = data[metric][uid]


    # we defined two additional metrics, 'warmup' and 'drain', which are defined
    # for all resources of the pilot.  `warmup` is defined as the time from
    # when the pilot becomes active, to the time the resource is first consumed
    # by a task.  `drain` is the inverse: the  time from when any task last
    # consumed the resource to the time when the pilot begins termination.
    for pilot in session.get(etype='pilot'):

      # if tdurations:
      #     # print('DEBUG: using tdurations')
      #     task_durations = tdurations
      # elif pilot.cfg['task_launch_method'] == 'PRTE':
      #     # print('DEBUG: using prte configuration')
      #     task_durations = TASK_DURATIONS_PRTE
      # else:
      #     # print('DEBUG: using default configuration')
      #     task_durations = TASK_DURATIONS_DEFAULT

        pt    = pilot.timestamps
        log.debug('timestamps:')
        for ts in pt():
            log.debug('    %10.2f  %-20s  %-15s  %-15s  %-15s  %-15s  %s',
                           ts[0],  ts[1], ts[2], ts[3], ts[4], ts[5], ts[6])

        p_min = None
        p_max = None

        try   : p_min = pilot.timestamps(event={1: 'bootstrap_0_start'})[0]
        except: pass

        try   : p_max = pilot.timestamps(event={1: 'bootstrap_0_stop'})[0]
        except: pass

        # fallback for missing bootstrap events
        if p_min is None: p_min = pilot.timestamps(state='PMGR_ACTIVE')
        if p_max is None: p_max = pilot.events[-1][ru.TIME]

        assert p_min is not None
        assert p_max is not None

        log.debug('pmin, pmax: %10.2f / %10.2f', p_min, p_max)

        pid  = pilot.uid
        rnd  = 'cores_per_node'
        rmap = 'core_map'

        if rtype == 'gpu':
            rnd  = 'gpus_per_node'
            rmap = 'gpu_map'

        pn = pilot.cfg['resource_details']['rm_info'][rnd]

        nodes, _, pnodes = _get_nodes(pilot)

        # find resource utilization scope for all resources. We begin filling
        # the resource dict with
        #
        #   resource_id : [t_min=None, t_max=None]
        #
        # and then iterate over all tasks.  Wen we find a task using some
        # resource id, we set or adjust t_min / t_max.
        resources = dict()
        for pnode in pnodes:
            idx = get_node_index(nodes, pnode['node_id'], pn)
            for c in range(idx[0], idx[1] + 1):
                resources[c] = [None, None]

        for task in session.get(etype='task'):

            if task.cfg.get('pilot') != pid:
                continue

            try:
                ranks  = task.cfg['slots']['ranks']
                tts    = task.timestamps
                task_min  = tts(event=task['consume']['exec_queue'][0]) [0]
                task_max  = tts(event=task['consume']['unschedule'][1])[-1]

            except:
                continue

            for rank in ranks:

                r0, _ = get_node_index(nodes, rank['node_id'], pn)

                for resource_map in rank[rmap]:
                    for resource in resource_map:
                        idx   = r0 + resource
                        t_min = resources[idx][0]
                        t_max = resources[idx][1]
                        if t_min is None or t_min > task_min: t_min = task_min
                        if t_max is None or t_max < task_max: t_max = task_max
                        resources[idx] = [t_min, t_max]

        # now sift through resources and find buckets of pairs with same t_min
        # or same t_max
        bucket_min  = dict()
        bucket_max  = dict()
        bucket_none = list()
        for idx in resources:

            t_min = resources[idx][0]
            t_max = resources[idx][1]

            if t_min is None:

                assert t_max is None
                bucket_none.append(idx)

            else:

                if t_min not in bucket_min:
                    bucket_min[t_min] = list()
                bucket_min[t_min].append(idx)

                if t_max not in bucket_max:
                    bucket_max[t_max] = list()
                bucket_max[t_max].append(idx)

        boxes_warm  = list()
        boxes_drain = list()
        boxes_idle  = list()

        # now cluster all lists and add the respective boxes
        for t_min in bucket_min:
            for r in cluster_resources(bucket_min[t_min]):
                boxes_warm.append([p_min, t_min, r[0], r[1]])

        for t_max in bucket_max:
            for r in cluster_resources(bucket_max[t_max]):
                boxes_drain.append([t_max, p_max, r[0], r[1]])

        for r in cluster_resources(bucket_none):
            boxes_idle.append([p_min, p_max, r[0], r[1]])

        if 'warm'  not in consumed: consumed['warm']  = dict()
        if 'drain' not in consumed: consumed['drain'] = dict()
        if 'idle'  not in consumed: consumed['idle']  = dict()

        consumed['warm'][pid]  = boxes_warm
        consumed['drain'][pid] = boxes_drain
        consumed['idle'][pid]  = boxes_idle

  # pprint.pprint(consumed)

    return consumed


# ------------------------------------------------------------------------------
#
def _get_nodes(pilot):

    pnodes = pilot.cfg['resource_details']['rm_info']['node_list']
    agents = pilot.cfg['resource_details']['rm_info'].get('agent_nodes', [])
    anodes = list()
    nodes  = list()

    for agent in agents:
        anodes.append(agents[agent])

    nodes = pnodes + anodes

    return nodes, anodes, pnodes


# ------------------------------------------------------------------------------
#
def _get_pilot_consumption(pilot, rtype):

    # Pilots consume resources in different ways:
    #
    #   - the pilot needs to bootstrap and initialize before becoming active,
    #     i.e., before it can begin to manage the workload, and needs to
    #     terminate and clean up during shutdown;
    #   - the pilot may block one or more nodes or cores for it's own components
    #     (sub-agents), and those components are not available for workload
    #     execution
    #   - the pilot may perform operations while managing the workload.
    #
    # This method will compute the first two contributions and part of the 3rd.
    # It will *not* account for those parts of the 3rd which are performed while
    # specfic resources are blocked for the affected workload element (task)
    # - those resource consumption is considered to be a consumption *of that
    # task*, which allows us to compute tasks specific resource utilization
    # overheads.

    pid = pilot.uid
    ret = dict()
    rnd = 'cores_per_node'

    if rtype == 'gpu':
        rnd  = 'gpus_per_node'

    pn = pilot.cfg['resource_details']['rm_info'][rnd]

    # Account for agent resources.  Agents use full nodes, i.e., cores and GPUs
    # We happen to know that agents use the first nodes in the allocation and
    # their resource tuples thus start at index `0`, but for completeness we
    # ensure that by inspecting the pilot cfg.

    # Duration is for all of the pilot runtime. This is not precises really,
    # since several bootstrapping phases happen before the agents exist - but we
    # consider the nodes blocked for the sub-agents from the get-go.
    t0, t1 = get_duration(pilot, PILOT_DURATIONS['agent']['total'])
    boxes  = list()

    # Substract agent nodes from the nodelist, so that we correctly attribute
    # other global pilot metrics to the remaining nodes.
    nodes, anodes, pnodes = _get_nodes(pilot)

    if anodes and t0 is not None:
        for anode in anodes:
            r0, r1 = get_node_index(nodes, anode['node_id'], pn)
            boxes.append([t0, t1, r0, r1])

    ret['agent'] = {pid: boxes}

    # account for all other pilot metrics
    for metric in PILOT_DURATIONS['consume']:

        if metric == 'idle':
            continue

        boxes = list()
        t0, t1 = get_duration(pilot, PILOT_DURATIONS['consume'][metric])

        if t0 is not None:
            for node in pnodes:
                r0, r1 = get_node_index(nodes, node['node_id'], pn)
                boxes.append([t0, t1, r0, r1])

        ret[metric] = {pid: boxes}

    return ret


# ------------------------------------------------------------------------------
#
def _get_task_consumption(session, task, rtype, tdurations=None):

    # we need to know what pilot the task ran on.  If we don't find a designated
    # pilot, no resources were consumed
    uid = task.uid
    pid = task.cfg['pilot']

    if not pid:
        return dict()

    # get the pilot for inspection
    pilot = session.get(uid=pid)

    if isinstance(pilot, list):
        assert len(pilot) == 1
        pilot = pilot[0]

    # FIXME: it is inefficient to query those values again and again
    nodes, _, _ = _get_nodes(pilot)

    rnd  = 'cores_per_node'
    rmap = 'core_map'

    if rtype == 'gpu':
        rnd  = 'gpus_per_node'
        rmap = 'gpu_map'

    pn = pilot.cfg['resource_details']['rm_info'][rnd]

    # Tasks consume only those resources they are scheduled on.
    if 'slots' not in task.cfg:
        return dict()

    ranks     = task.cfg['slots']['ranks']
    resources = list()
    for rank in ranks:

        r0, _ = get_node_index(nodes, rank['node_id'], pn)

        for resource_map in rank[rmap]:
            for resource in resource_map:
                resources.append(r0 + resource)

    # find continuous stretched of resources to minimize number of boxes
    resources = cluster_resources(resources)

    if tdurations:
        task_durations = tdurations
    else:
        task_durations = TASK_DURATIONS_DEFAULT

    if _debug:
        print()

    ret = dict()
    for metric in task_durations['consume']:

        boxes = list()
        t0, t1 = get_duration(task, task_durations['consume'][metric])


        if t0 is not None:

            if _debug:
                print('%s: %-15s : %10.3f - %10.3f = %10.3f'
                     % (task.uid, metric, t1, t0, t1 - t0))
            for r in resources:
                boxes.append([t0, t1, r[0], r[1]])

        else:
            if _debug:
                print('%s: %-15s : -------------- ' % (task.uid, metric))
                dur = task_durations['consume'][metric]
                print(dur)

                for e in dur:
                    if ru.STATE in e and ru.EVENT not in e:
                        e[ru.EVENT] = 'state'

                t0 = task.timestamps(event=dur[0])
                t1 = task.timestamps(event=dur[1])
                print(t0)
                print(t1)
                for e in task.events:
                    print('\t'.join([str(x) for x in e]))

              # sys.exit()

        ret[metric] = {uid: boxes}

    return ret


# ------------------------------------------------------------------------------
#
def get_resource_transitions(pilot, task_metrics=None, pilot_metrics=None):

    task_metrics  = task_metrics  or TASK_DURATIONS_DEFAULT
    pilot_metrics = pilot_metrics or PILOT_DURATIONS


    # we try to find points in time where resource usage moved from purpose A to
    # purpose B.  For example, consider this metric:
    #
    #   'exec_queue'  : [{ru.EVENT: 'schedule_ok'    },
    #                    {ru.STATE: s.AGENT_EXECUTING}],
    #   'exec_prep'   : [{ru.STATE: s.AGENT_EXECUTING},
    #                    {ru.EVENT: 'exec_start'     }],
    #   'exec_cmd'    : [{ru.EVENT: 'exec_start'     },
    #                    {ru.EVENT: 't_start'        }],
    #
    # then we convert this into the following structure:
    #
    # [
    #   # event                          from         to
    #   [{ru.EVENT: 'schedule_ok'    },  None,        'exec_queue'],
    #   [{ru.STATE: s.AGENT_EXECUTING}, 'exec_queue', 'exec_prep'],
    #   [{ru.EVENT: 'exec_start'     }, 'exec_prep',  'exec_rp'],
    #   [{ru.EVENT: 't_start'        }, 'exec_rp',    'None]
    # ]
    #
    # which we will use like this:
    #
    #   - go through all events for a task
    #   - if event is in the list above
    #     - reduce resources in `from` metric
    #     - increase resource in `to` metric
    #
    # If `from` or `to` are None, then the resources are taken from / given to
    # the pilot's idle resources.  We thus rename the 'None' to 'idle'.

    # dig though metric, find all pairs of matching start/stop events
    task_transitions = list()
    for metric,spec in task_metrics['consume'].items():

        # find the metric's transition spec
        m = None
        for m in task_transitions:
            if m[0] == spec[0]:
                break
            m = None

        # if we don't find that spec registered, register it as start event
        if not m:
            task_transitions.append([spec[0], None, metric])
        else:
            # if we know the transition, just register the stop event
            assert m[2] is None
            m[2] = metric

        # the inverse of the above where we check stop events, register, and
        # insert start events (in case transitions are not causally ordered)
        m = None
        for m in task_transitions:
            if m[0] == spec[1]:
                break
            m = None
        if not m:
            task_transitions.append([spec[1], metric, None])
        else:
            assert m[1] is None
            m[1] = metric

    # task transitions which, after the above search, miss start or stop events
    # will add / remove resources from the pilot's idle pool.
    for t in task_transitions:
        if t[1] is None: t[1] = 'idle'
        if t[2] is None: t[2] = 'idle'

    # do the same for the pilot metrics / transitions, only that now we
    # translate `None` as `system`, which is where the pilot obtains it's
    # resources from.  Also note that some transition events can have double
    # entries, for example, if a event passes some resources to a sub-agent, and
    # some other resources remain in the pilot.
    #
    # Note that we also handle the `agent` metrics to cover resources used by
    # the agent (and thus not making it into usable).
    pilot_transitions = list()
    for metric,spec in pilot_metrics['consume'].items():

        # same as above, but for pilot transition events
        m = None
        for m in pilot_transitions:
            if m[0] == spec[0]:
                break
            m = None
        if not m:
            pilot_transitions.append([spec[0], None, metric])
        else:
            assert m[2] is None
            m[2] = metric

        m = None
        for m in pilot_transitions:
            if m[0] == spec[1]:
                break
            m = None
        if not m:
            pilot_transitions.append([spec[1], metric, None])
        else:
            assert m[1] is None
            m[1] = metric

    for t in pilot_transitions:
        # agent resources are taked from the pilot
        # pilot resources are taken from the system
        if t[1] is None:
            if t[2] == 'agent': t[1] = 'setup_1'
            else              : t[1] = 'system'
        if t[2] is None:
            if t[1] == 'agent': t[2] = 'term'
            else              : t[2] = 'system'

    return pilot_transitions, task_transitions


# ------------------------------------------------------------------------------
#
def get_resource_timelines(task, transitions):
    '''
    For each specific task, return a set of tuples of the form:

        [start, stop, metric]

    which reports what metric has been used during what time span.
    '''

    # we need to know what pilot the task ran on.  If we don't find a designated
    # pilot, no resources were consumed
    tid = task.uid
    pid = task.cfg['pilot']

    if not pid:
        # task was never assigned to a pilot
        return dict()

    if 'slots' not in task.cfg:
        # the task was never scheduled
        return dict()

    ret = list()
    for metric, spec in transitions['consume'].items():
        t0, t1 = get_duration(task, spec)
        ret.append([t0, t1, metric, tid])

    return ret


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2016-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import multiprocessing

import math

from .base import RMInfo, ResourceManager


# ------------------------------------------------------------------------------
#
class Fork(ResourceManager):

    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, rm_info: RMInfo) -> RMInfo:

        # check if the requested cores are available
        detected_cores = multiprocessing.cpu_count()

        if detected_cores != rm_info.requested_cores:
            if self._cfg.resource_cfg.fake_resources:
                self._log.info(
                    'using %d requested cores instead of %d available cores',
                    rm_info.requested_cores, detected_cores)
            else:
                if detected_cores < rm_info.requested_cores:
                    raise RuntimeError('insufficient cores found (%d < %d)' %
                                       (detected_cores, rm_info.requested_cores))

        if not rm_info.requested_nodes:
            cores = rm_info.requested_cores
            cpn   = rm_info.cores_per_node
            rm_info.requested_nodes = math.ceil(cores / cpn)

        # FIXME: number of GPUs should also be checked - how? Should we use
        #        hwlock if available?

        nodes = [('localhost', rm_info.cores_per_node)
                 for idx in range(rm_info.requested_nodes)]

        rm_info.node_list = self._get_node_list(nodes, rm_info)

        # UIDs need to be made unique
        for idx, node in enumerate(rm_info.node_list):
            node['node_id'] = '%s_%04d' % (node['node_name'], idx)

        return rm_info


# ------------------------------------------------------------------------------





import os
import sys
import copy
import time

from typing import Dict, Union

import threading         as mt

import radical.utils     as ru

from .. import utils     as rpu
from .. import states    as rps
from .. import constants as rpc

from .. import Session, Task, TaskDescription, TASK_EXECUTABLE

from ..task_description import RAPTOR_WORKER


def out(msg):
    sys.stdout.write('%s\n' % msg)
    sys.stdout.flush()


# ------------------------------------------------------------------------------
#
class Master(rpu.Component):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg=None):

        self._uid      = os.environ['RP_TASK_ID']
        self._sid      = os.environ['RP_SESSION_ID']
        self._name     = os.environ['RP_TASK_NAME']
        self._sbox     = os.environ['RP_TASK_SANDBOX']
        self._psbox    = os.environ['RP_PILOT_SANDBOX']

        self._workers  = dict()      # wid: worker
        self._tasks    = dict()      # bookkeeping of submitted requests
        self._lock     = mt.Lock()   # lock the request dist on updates
        self._term     = mt.Event()  # termination signal
        self._thread   = None        # run loop

        cfg            = self._get_config(cfg)
        self._session  = Session(cfg=cfg, uid=cfg.sid, _primary=False)

        rpu.Component.__init__(self, cfg, self._session)

        self.register_publisher(rpc.STATE_PUBSUB, self._psbox)
        self.register_publisher(rpc.CONTROL_PUBSUB, self._psbox)

        self.register_subscriber(rpc.STATE_PUBSUB,   self._state_cb, self._psbox)
        self.register_subscriber(rpc.CONTROL_PUBSUB, self._control_cb, self._psbox)

        # send new worker tasks and agent input staging / agent scheduler
        self.register_output(rps.AGENT_STAGING_INPUT_PENDING,
                             rpc.AGENT_STAGING_INPUT_QUEUE, self._psbox)

        # set up zmq queues between the agent scheduler and this master so that
        # we can receive new requests from RP tasks
        qname = '%s.input_queue' % self._uid
        input_cfg = ru.Config(cfg={'channel'   : qname,
                                   'type'      : 'queue',
                                   'uid'       : '%s_input' % self._uid,
                                   'path'      : self._sbox,
                                   'stall_hwm' : 0,
                                   'bulk_size' : 1})

        self._input_queue = ru.zmq.Queue(input_cfg)
        self._input_queue.start()

        # send completed request tasks to agent output staging / tmgr
        self.register_output(rps.AGENT_STAGING_OUTPUT_PENDING,
                             rpc.AGENT_STAGING_OUTPUT_QUEUE, self._psbox)

        # set up zmq queues between this master and all workers for request
        # distribution and result collection
        req_cfg = ru.Config(cfg={'channel'   : 'raptor_tasks',
                                 'type'      : 'queue',
                                 'uid'       : self._uid + '.req',
                                 'path'      : self._sbox,
                                 'stall_hwm' : 0,
                                 'bulk_size' : 1})

        res_cfg = ru.Config(cfg={'channel'   : 'raptor_results',
                                 'type'      : 'queue',
                                 'uid'       : self._uid + '.res',
                                 'path'      : self._sbox,
                                 'stall_hwm' : 0,
                                 'bulk_size' : 1})

        self._req_queue = ru.zmq.Queue(req_cfg)
        self._res_queue = ru.zmq.Queue(res_cfg)

        self._req_queue.start()
        self._res_queue.start()

        self._req_addr_put = str(self._req_queue.addr_put)
        self._req_addr_get = str(self._req_queue.addr_get)

        self._res_addr_put = str(self._res_queue.addr_put)
        self._res_addr_get = str(self._res_queue.addr_get)

        # this master will put requests onto the request queue, and will get
        # responses from the response queue.  Note that the responses will be
        # delivered via an async callback (`self._result_cb`).
        self._req_put = ru.zmq.Putter('raptor_tasks',    self._req_addr_put)
        self._res_get = ru.zmq.Getter('raptor_results',  self._res_addr_get,
                                                         cb=self._result_cb)

        # for the workers it is the opposite: they will get requests from the
        # request queue, and will send responses to the response queue.
        self._info = {'req_addr_get': self._req_addr_get,
                      'res_addr_put': self._res_addr_put}

        # make sure the channels are up before allowing to submit requests
        time.sleep(1)

        # begin to receive tasks in that queue
        ru.zmq.Getter(qname, self._input_queue.addr_get, cb=self._request_cb)

        # and register that input queue with the scheduler
        self._log.debug('registered raptor queue')
        self.publish(rpc.CONTROL_PUBSUB,
                      {'cmd': 'register_raptor_queue',
                       'arg': {'name' : self._uid,
                               'queue': qname,
                               'addr' : str(self._input_queue.addr_put)}})

        # also create a ZMQ server endpoint for the workers to
        # send task execution requests back to the master
        self._task_service = ru.zmq.Server()
        self._task_service.register_request('run_task', self._run_task)
        self._task_service.start()
        self._task_service_data = dict()   # task.uid : [mt.Event, task]

        # all comm channels are set up - begin to work
        self._log.debug('startup complete')


    # --------------------------------------------------------------------------
    #
    def _get_config(self, cfg=None):
        '''
        derive a worker base configuration from the control pubsub configuration
        '''

        # FIXME: use registry for comm EP info exchange, not cfg files

        if cfg is None:
            cfg = dict()

        if cfg and 'path' in cfg:
            del cfg['path']

        ru.dict_merge(cfg, ru.read_json('%s/control_pubsub.json' % self._psbox))

        del cfg['channel']
        del cfg['cmgr']

        cfg['log_lvl'] = 'warn'
        cfg['kind']    = 'master'
        cfg['sid']     = self._sid
        cfg['base']    = self._sbox
        cfg['path']    = self._sbox

        cfg = ru.Config(cfg=cfg)
        cfg['uid'] = self._uid

        return cfg


    # --------------------------------------------------------------------------
    #
    @property
    def workers(self):
        return self._workers


    # --------------------------------------------------------------------------
    #
    def _control_cb(self, topic, msg):

        cmd = msg['cmd']
        arg = msg['arg']

        if cmd == 'worker_register':

            uid  = arg['uid']
            info = arg['info']

            self._log.debug('register %s', uid)

            with self._lock:
                if 'uid' not in self._workers:
                    return
                self._workers[uid]['info']  = info
                self._workers[uid]['state'] = 'ACTIVE'

            self._log.debug('info: %s', info)


        elif cmd == 'worker_unregister':

            uid = arg['uid']
            self._log.debug('unregister %s', uid)

            with self._lock:
                if 'uid' not in self._workers:
                    return
                self._workers[uid]['status'] = 'DONE'


    # --------------------------------------------------------------------------
    #
    def _state_cb(self, topic, msg):

        cmd = msg['cmd']
        arg = msg['arg']

        if cmd == 'raptor_state_update':

            tasks = ru.as_list(arg)

            for task in tasks:
                uid   = task['uid']
                state = task['state']

                if uid in self._task_service_data:
                    # update task info and signal task service thread
                    self._log.debug('unlock 2 %s', uid)
                    self._task_service_data[uid][1] = task
                    self._task_service_data[uid][0].set()


        elif cmd == 'update':

            for thing in ru.as_list(arg):

                uid   = thing['uid']
                state = thing['state']

                if uid in self._workers:
                    if state == rps.AGENT_STAGING_OUTPUT:
                        with self._lock:
                            self._workers[uid]['state'] = 'DONE'

            self.state_cb(ru.as_list(arg))

        return True


    # --------------------------------------------------------------------------
    #
    def state_cb(self, tasks):

        pass


    # --------------------------------------------------------------------------
    #
    def submit_workers(self, descr: Dict[str, Union[str, int]],
                             count: int) -> None:
        '''
        Submit`count` workers per given `descr`, and pass the queue info as
        configuration file.  Do *not* wait for the workers to come up - they are
        expected to register via the control channel.

        The `descr` dict is expected to support the following keys:

          - named_env       : environment to use (same as master usually)
          - ranks           : number of MPI ranks per worker
          - cores_per_rank  : int, number of cores per worker rank
          - gpus_per_rank   : int, number of gpus per worker rank
          - worker_class    : str, type name of worker class to execute
          - worker_file     : str, optional if an RP worker class is used

        Note that only one rank (presumably rank 0) should register with the
        master - the worker ranks are expected to syncronize their ranks as
        needed.
        '''
        with self._lock:

            tasks = list()
            base  = self._uid

            cfg            = copy.deepcopy(self._cfg)
            cfg['descr']   = descr
            cfg['info']    = self._info
            cfg['ts_addr'] = self._task_service.addr

            for i in range(count):

                uid        = '%s.worker.%04d' % (base, i)
                cfg['uid'] = uid

                fname = './%s.json' % uid
                ru.write_json(cfg, fname)

                td = dict()
                td['mode']             = RAPTOR_WORKER
                td['named_env']        = descr.get('named_env')
                td['ranks']            = descr['ranks']
                td['threading_type']   = rpc.POSIX
                td['cores_per_rank']   = descr.get('cores_per_rank', 1)
                td['gpus_per_rank']    = descr.get('gpus_per_rank', 0)
                td['environment']      = descr.get('environment', {})

                # this master is obviously running in a suitable python3 env,
                # so we expect that the same env is also suitable for the worker
                # NOTE: shell escaping is a bit tricky here - careful on change!
                td['pre_exec']   = descr.get('pre_exec', [])
                td['executable'] = 'python3'
                td['arguments']  = [
                        '-c',
                        'import radical.pilot as rp; '
                        "rp.raptor.Worker.run('%s', '%s', '%s')"
                            % (descr.get('worker_file', ''),
                               descr.get('worker_class', 'DefaultWorker'),
                               fname)]



                # all workers run in the same sandbox as the master
                task = dict()

                task['description']       = TaskDescription(td).as_dict()
                task['state']             = rps.AGENT_STAGING_INPUT_PENDING
                task['status']            = 'NEW'
                task['type']              = 'task'
                task['uid']               = uid
                task['task_sandbox_path'] = self._sbox
                task['task_sandbox']      = 'file://localhost/' + self._sbox
                task['pilot_sandbox']     = os.environ['RP_PILOT_SANDBOX']
                task['session_sandbox']   = os.environ['RP_SESSION_SANDBOX']
                task['resource_sandbox']  = os.environ['RP_RESOURCE_SANDBOX']
                task['pilot']             = os.environ['RP_PILOT_ID']
                task['resources']         = {'cpu': td['ranks'] *
                                                    td.get('cores_per_rank', 1),
                                             'gpu': td['ranks'] *
                                                    td.get('gpus_per_rank', 1)}
                tasks.append(task)

                # NOTE: the order of insert / state update relies on that order
                #       being maintained through the component's message push,
                #       the update worker's message receive up to the insertion
                #       order into the update worker's DB bulk op.
                self._log.debug('insert %s', uid)
                self.publish(rpc.STATE_PUBSUB, {'cmd': 'insert', 'arg': task})

                self._workers[uid] = dict()
                self._workers[uid]['state'] = 'NEW'

            self.advance(tasks, publish=True, push=True)


    # --------------------------------------------------------------------------
    #
    def wait(self, count=None, uids=None):
        '''
        wait for `n` workers, *or* for workers with given UID, *or* for all
        workers to become available, then return.
        '''

        if not count and not uids:
            uids = list(self._workers.keys())

        if count:
            self._log.debug('wait for %d workers', count)
            while True:
                stats = {'NEW'    : 0,
                         'ACTIVE' : 0,
                         'DONE'   : 0,
                         'FAILED' : 0}

                with self._lock:
                    for w in self._workers.values():
                        stats[w['status']] += 1

                self._log.debug('stats: %s', stats)
                n = stats['ACTIVE'] + stats['DONE'] + stats['FAILED']
                if n >= count:
                    self._log.debug('wait ok')
                    return
                time.sleep(1)

        elif uids:
            self._log.debug('wait for workers: %s', uids)
            while True:
                with self._lock:
                    stats = [self._workers[uid]['status'] for uid in uids]
                n = stats['ACTIVE'] + stats['DONE'] + stats['FAILED']
                self._log.debug('stats [%d]: %s', n, stats)
                if n == len(uids):
                    self._log.debug('wait ok')
                    return
                time.sleep(1)


    # --------------------------------------------------------------------------
    #
    def start(self):

        self._thread = mt.Thread(target=self._run)
        self._thread.daemon = True
        self._thread.start()


    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._log.debug('set term from stop: %s', ru.get_stacktrace())
        self._term.set()

        rpu.Component.stop(self)

        self.terminate()


    # --------------------------------------------------------------------------
    #
    def alive(self):

        if not self._thread or self._term.is_set():
            return False
        return True


    # --------------------------------------------------------------------------
    #
    def join(self):

        if self._thread:
            self._thread.join()


    # --------------------------------------------------------------------------
    #
    def _run(self):

        # wait for the submitted requests to complete
        while not self._term.wait(timeout=60):

            self._log.debug('still alive')

        self._log.debug('terminate')


    # --------------------------------------------------------------------------
    #
    def _run_task(self, td):
        '''
        accept a single task request for execution, execute it and wait for it's
        completion before returning the call.

        Note: this call is running in a separate thread created by an ZMQ
              Server instance and will thus not block the master's progress.
        '''

        # we get a dict but want a proper `TaskDescription` instance
        td = TaskDescription(td)

        # Create a new task ID for the submitted task (we do not allow
        # user-specified IDs in this case as we can't check uniqueness with the
        # client tmgr).  Then submit that task and wait for the `result_cb` to
        # report completion of the task

        tid   = '%s.%s' % (self.uid, ru.generate_id('subtask'))
        event = mt.Event()

        td['uid'] = tid

        self._task_service_data[tid] = [event, td]
        self.submit_tasks([td])

        # wait for the result cb to pick up the td again
        event.wait()

        # update td info and remove data
        task = self._task_service_data[tid][1]
        del self._task_service_data[tid]

        # the task is completed and we can return it to the caller
        return task


    # --------------------------------------------------------------------------
    #
    def submit_tasks(self, tasks):
        '''
        submit a list of tasks to the task queue
        We expect to get either `TaskDescription` instances which will then get
        converted into task dictionaries and pushed out, or we get task
        dictionaries which are used as is.
        '''

        raptor_tasks     = list()
        executable_tasks = list()
        for task in ru.as_list(tasks):

            if isinstance(task, TaskDescription):
                # convert to task dict
                task = Task(self, task, origin='raptor').as_dict()

            assert 'description' in task

            mode = task['description'].get('mode', TASK_EXECUTABLE)
            if mode == TASK_EXECUTABLE:
                executable_tasks.append(task)
            else:
                raptor_tasks.append(task)

        self._submit_executable_tasks(executable_tasks)
        self._submit_raptor_tasks(raptor_tasks)


    # --------------------------------------------------------------------------
    #
    def _submit_raptor_tasks(self, tasks):

        if tasks:

            self.advance(tasks, state=rps.AGENT_EXECUTING,
                                publish=True, push=False)

            self._req_put.put(tasks)


    # --------------------------------------------------------------------------
    #
    def _submit_executable_tasks(self, tasks):
        '''
        Submit tasks per given task description to the agent this
        master is running in.
        '''

        if not tasks:
            return

        for task in tasks:

            td = task['description']

            sbox = '%s/%s' % (self._sbox, td['uid'])

          # task['uid']               = td.get('uid')
            task['state']             = rps.AGENT_STAGING_INPUT_PENDING
            task['task_sandbox_path'] = sbox
            task['task_sandbox']      = 'file://localhost/' + sbox
            task['pilot_sandbox']     = os.environ['RP_PILOT_SANDBOX']
            task['session_sandbox']   = os.environ['RP_SESSION_SANDBOX']
            task['resource_sandbox']  = os.environ['RP_RESOURCE_SANDBOX']
            task['pilot']             = os.environ['RP_PILOT_ID']
            task['resources']         = {'cpu': td['ranks'] *
                                                td.get('cores_per_rank', 1),
                                         'gpu': td['ranks'] *
                                                td.get('gpus_per_rank', 0)}

            # NOTE: the order of insert / state update relies on that order
            #       being maintained through the component's message push,
            #       the update worker's message receive up to the insertion
            #       order into the update worker's DB bulk op.
            self._log.debug('insert %s', td['uid'])
            self.publish(rpc.STATE_PUBSUB, {'cmd': 'insert', 'arg': task})

        self.advance(tasks, state=rps.AGENT_STAGING_INPUT_PENDING,
                            publish=True, push=True)


    # --------------------------------------------------------------------------
    #
    def _request_cb(self, tasks):

        tasks = ru.as_list(tasks)

        self._log.debug('request_cb: %d', len(tasks))

        try:
            filtered = self.request_cb(tasks)
            if filtered:
                for task in filtered:
                    self._log.debug('REQ cb: %s' % task['uid'])
                self.submit_tasks(filtered)

        except:
            self._log.exception('request cb failed')
            # FIXME: fail tasks


    # --------------------------------------------------------------------------
    #
    def request_cb(self, tasks):

        # FIXME: document task format
        return tasks


    # --------------------------------------------------------------------------
    #
    def _result_cb(self, tasks):

        tasks = ru.as_list(tasks)

        try:
            self.result_cb(tasks)

        except:
            self._log.exception('result callback failed')

        for task in tasks:

            # check if the task was submited via the task_service EP
            tid = task['uid']

            if tid in self._task_service_data:
                # update task info and signal task service thread
                self._log.debug('unlock 1 %s', tid)
                self._task_service_data[tid][1] = task
                self._task_service_data[tid][0].set()

            ret = int(task.get('exit_code', -1))

            if ret == 0: task['target_state'] = rps.DONE
            else       : task['target_state'] = rps.FAILED

        self.advance(tasks, rps.AGENT_STAGING_OUTPUT_PENDING,
                            publish=True, push=True)


    # --------------------------------------------------------------------------
    #
    def result_cb(self, tasks):

        # FIXME: document task format

        pass


    # --------------------------------------------------------------------------
    #
    def terminate(self):
        '''
        terminate all workers
        '''

        # unregister input queue
        self.publish(rpc.CONTROL_PUBSUB, {
                    'cmd': 'unregister_raptor_queue',
                    'arg': {'name' : self._uid,
                            'queue': self._input_queue.channel}})

        self._log.debug('set term from terminate')
        self._term.set()
        for uid in self._workers:
            self.publish(rpc.CONTROL_PUBSUB, {'cmd': 'worker_terminate',
                                              'arg': {'uid': uid}})

        # wait for workers to terminate
      # uids = self._workers.keys()
      # FIXME TS
      # while True:
      #     states = [self._workers[uid]['state'] for uid in uids]
      #     if set(states) == {'DONE'}:
      #         break
      #     self._log.debug('term states: %s', states)
      #     time.sleep(1)

        self._log.debug('all workers terminated')


# ------------------------------------------------------------------------------





import time
import random

import radical.pilot as rp


# ------------------------------------------------------------------------------
#
class MyWorker(rp.raptor.MPIWorker):
    '''
    This class provides the required functionality to execute work requests.
    In this simple example, the worker only implements a single call: `hello`.
    '''


    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg):

        rp.raptor.MPIWorker.__init__(self, cfg)


    # --------------------------------------------------------------------------
    #
    def my_hello(self, uid, count=0):
        '''
        important work
        '''

        self._prof.prof('app_start', uid=uid)

        out = 'hello %5d @ %.2f [%s]' % (count, time.time(), self._uid)
        time.sleep(random.randint(1, 5))

        self._log.debug(out)

        self._prof.prof('app_stop', uid=uid)
        self._prof.flush()

        td = rp.TaskDescription({
                'mode'            : rp.TASK_EXECUTABLE,
                'scheduler'       : None,
                'ranks'           : 2,
                'executable'      : '/bin/sh',
                'arguments'       : ['-c',
                            'echo "hello $RP_RANK/$RP_RANKS: $RP_TASK_ID"']})

        td = rp.TaskDescription({
              # 'uid'             : 'task.call.w.000000',
              # 'timeout'         : 10,
                'mode'            : rp.TASK_EXECUTABLE,
                'ranks'           : 2,
                'executable'      : 'radical-pilot-hello.sh',
                'arguments'       : ['1', 'task.call.w.000000']})

        master = self.get_master()
        task   = master.run_task(td)

        print(task['stdout'])


# ------------------------------------------------------------------------------





# pylint: disable=access-member-before-definition

__copyright__ = 'Copyright 2013-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import radical.utils as ru

# task modes
TASK_EXECUTABLE  = 'task.executable'
TASK_FUNCTION    = 'task.function'
TASK_METHOD      = 'task.method'
TASK_EVAL        = 'task.eval'
TASK_EXEC        = 'task.exec'
TASK_PROC        = 'task.proc'
TASK_SHELL       = 'task.shell'
RAPTOR_MASTER    = 'raptor.master'
RAPTOR_WORKER    = 'raptor.worker'

# task description attributes
UID              = 'uid'
NAME             = 'name'
MODE             = 'mode'

# mode: TASK_EXECUTABLE
EXECUTABLE       = 'executable'
ARGUMENTS        = 'arguments'

# mode: TASK_METHOD  # FIXME
METHOD           = 'method'
ARGS             = 'args'
KWARGS           = 'kwargs'

# mode: TASK_FUNCTION
FUNCTION         = 'function'
ARGS             = 'args'
KWARGS           = 'kwargs'

# mode: TASK_EXEC, TASK_EVAL
CODE             = 'code'

# mode: TASK_PROC, TASK_SHELL
COMMAND          = 'command'

# environment
ENVIRONMENT      = 'environment'
NAMED_ENV        = 'named_env'
SANDBOX          = 'sandbox'

# resource requirements
RANKS            = 'ranks'                    # ranks
CORES_PER_RANK   = 'cores_per_rank'           # cores per rank
GPUS_PER_RANK    = 'gpus_per_rank'            # gpus per rank
THREADING_TYPE   = 'threading_type'           # OpenMP?
GPU_TYPE         = 'gpu_type'                 # CUDA / ROCm?
LFS_PER_RANK     = 'lfs_per_rank'             # disk space per rank
MEM_PER_RANK     = 'mem_per_rank'             # memory per rank

# deprecated
CPU_PROCESSES    = 'cpu_processes'            # ranks
CPU_PROCESS_TYPE = 'cpu_process_type'         # disk space per rank
CPU_THREADS      = 'cpu_threads'              # memory per rank
CPU_THREAD_TYPE  = 'cpu_thread_type'          # OpenMP?

GPU_PROCESSES    = 'gpu_processes'            # gpus per rank
GPU_PROCESS_TYPE = 'gpu_process_type'         # CUDA?
GPU_THREADS      = 'gpu_threads'              # part of gpu?
GPU_THREAD_TYPE  = 'gpu_thread_type'          # ?

LFS_PER_PROCESS  = 'lfs_per_process'          # disk space per rank
MEM_PER_PROCESS  = 'mem_per_process'          # memory per rank

# task setup
INPUT_STAGING    = 'input_staging'
OUTPUT_STAGING   = 'output_staging'
STAGE_ON_ERROR   = 'stage_on_error'
PRE_LAUNCH       = 'pre_launch'
PRE_EXEC         = 'pre_exec'
PRE_RANK         = 'pre_rank'
POST_LAUNCH      = 'post_launch'
POST_EXEC        = 'post_exec'
POST_RANK        = 'post_rank'
CLEANUP          = 'cleanup'
PILOT            = 'pilot'
STDOUT           = 'stdout'
STDERR           = 'stderr'
RESTARTABLE      = 'restartable'
SCHEDULER        = 'scheduler'
TAGS             = 'tags'
METADATA         = 'metadata'


# ------------------------------------------------------------------------------
#
class TaskDescription(ru.TypedDict):
    """
    A TaskDescription object describes the requirements and properties
    of a :class:`radical.pilot.Task` and is passed as a parameter to
    :meth:`radical.pilot.TaskManager.submit_tasks` to instantiate and run
    a new task.

    .. note:: A TaskDescription **MUST** define at least an
              `executable` -- all other elements are optional.

    .. data:: uid

       [type: `str` | default: `""`] A unique ID for the task. This attribute
       is optional, a unique ID will be assigned by RP if the field is not set.

    .. data:: name

       [type: `str` | default: `""`] A descriptive name for the task. This
       attribute can be used to map individual tasks back to application level
       workloads.

    .. data:: mode

       [type: `str` | default: `"executable"`] The execution mode to be used for
       this task.  The following modes are accepted:

         - TASK_EXECUTABLE: the task is spawned as an external executable via a
           resource specific launch method (srun, aprun, mpiexec, etc).
           required attributes: `executable`
           related  attributes: `arguments`

         - TASK_FUNCTION: the task references a python function to be called.
           required attributes: `function`
           related  attributes: `args`
           related  attributes: `kwargs`

         - TASK_METHOD: the task references a raptor worker method to be
           called.
           required attributes: `method`
           related  attributes: `args`
           related  attributes: `kwargs`

         - TASK_EVAL: the task is a code snippet to be evaluated.
           required attributes: `code`

         - TASK_EXEC: the task is a code snippet to be `exec`'ed.
           required attributes: `code`

         - TASK_SHELL: the task is a shell command line to be run.
           required attributes: `command`

         - TASK_PROC: the task is a single core process to be executed.
           required attributes: `executable`
           related  attributes: `arguments`

        There exists a certain overlap between `TASK_EXECUTABLE`, `TASK_SHELL`
        and `TASK_PROC` modes.  As a general rule, `TASK_SHELL` and `TASK_PROC`
        should be used for short running tasks which require a single core and
        no additional resources (gpus, storage, memory).  `TASK_EXECUTABLE`
        should be used for all other tasks and is in fact the default.
        `TASK_SHELL` should only be used if the command to be run requires shell
        specific functionality (pipes, I/O redirection) which cannot easily be
        mapped to other task attributes.


    .. data:: executable

       [type: `str` | default: `""`] The executable to launch. The executable
       is expected to be either available via `$PATH` on the target resource,
       or to be an absolute path.

    .. data:: arguments

       [type: `list` | default: `[]`] The command line arguments for the given
       `executable` (`list` of `strings`).

    .. data:: code

       [type: `str` | default: `""`] The code to run.  This field is expected to
       contain valid python code which is executed when the task mode is
       `TASK_EXEC` or `TASK_EVAL`.

    .. data:: function

       [type: `str` | default: `""`] The function to run.  This field is
       expected to contain a python function name which can be resolved in the
       scope of the respective RP worker implementation (see documentation
       there).  The task mode must be set to `TASK_FUNCTION`.  `args` and
       `kwargs` are passed as function parameters.

    .. data:: args

       [type: `list` | default: `[]`] Unnamed arguments to be passed to the
       `function` (see above).  This field will be serialized  with `msgpack`
       and can thus contain any serializable data types.

    .. data:: kwargs

       [type: `dict` | default: `{}`] Named arguments to be passed to the
       `function` (see above).  This field will be serialized  with `msgpack`
       and can thus contain any serializable data types.

    .. data:: command

       [type: `str` | default: `""`] A shell command to be executed.  This
       attribute is used for the `TASK_SHELL` mode.

    .. data:: ranks

       [type: `int` | default: `1`] The number of application processes to start
       on CPU cores.  For two ranks or more, an MPI communicator will be
       available to the processes.

       `ranks` replaces the deprecated attribute `cpu_processes`.  The attribute
       `cpu_process_type` was previously used to signal the need for an MPI
       communicator - that attribute is now also deprecated and will be ignored.

    .. data:: cores_per_rank

       [type: `int` | default: `1`] The number of cpu cores each process will
       have available to start it's own threads or processes on.  By default,
       `core` refers to an physical CPU core - but if the pilot has been
       launched with SMT-settings > 1, `core` will refer to a virtual core or
       hyperthread instead (the name depends on the CPU vendor).

       `cores_per_rank` replaces the deprecated attribute `cpu_threads`.

    .. data:: threading_type

       [type: `str` | default: `""`] The thread type, influences startup and
       environment (`<empty>/POSIX`, `OpenMP`).

       `threading_type` replaces the deprecated attribute `cpu_thread_type`.

    .. data:: gpus_per_rank

       [type: `int` | default: `0`] The number of gpus made available to each
       rank.

       `gpus_per_rank` replaces the deprecated attribute `gpu_processes`.  The
       attributes `gpu_threads` and `gpu_process_type` are also deprecated and
       will be ignored.

    .. data:: gpu_type

       [type: `str` | default: `""`] The type of GPU environment to provide to
       the ranks (`<empty>`, `CUDA`, `ROCm`).

       `gpu_type` replaces the deprecated attribute `gpu_thread_type`.


    .. data:: lfs_per_rank

       [type: `int` | default: `0`] Local File Storage per rank - amount of
       data (MB) required on the local file system of the node.

       `lfs_per_rank` replaces the deprecated attribute `lfs_per_process`.

    .. data:: mem_per_rank

       [type: `int` | default: `0`] Amount of physical memory required per
       rank.

       `mem_per_rank` replaces the deprecated attribute `mem_per_process`.

    .. data:: environment

       [type: `dict` | default: `{}`] Environment variables to set in the
       environment before the execution (launching picked `LaunchMethod`).

    .. data:: named_env

       [type: `str` | default: `""`] A named virtual environment as prepared by
       the pilot. The task will remain in `AGENT_SCHEDULING` state until that
       environment gets created.

    .. data:: sandbox

       [type: `str` | default: `""`] This specifies the working directory of
       the task. That directory *MUST* be relative to the pilot sandbox. It
       will be created if it does not exist. By default, the sandbox has
       the name of the task's uid.

    .. data:: stdout

       [type: `str` | default: `""`] The name of the file to store stdout. If
       not set then the name of the following format will be used: `<uid>.out`.

    .. data:: stderr

       [type: `str` | default: `""`] The name of the file to store stderr. If
       not set then the name of the following format will be used: `<uid>.err`.

    .. data:: input_staging

       [type: `list` | default: `[]`] The files that need to be staged before
       the execution (`list` of `staging directives`, see below).

    .. data:: output_staging

       [type: `list` | default: `[]`] The files that need to be staged after
       the execution (`list` of `staging directives`, see below).

    .. data:: stage_on_error

       [type: `bool` | default: `False`] Output staging is normally skipped on
       `FAILED` or `CANCELED` tasks, but if this flag is set, staging is
       attempted either way. This may though lead to additional errors if the
       tasks did not manage to produce the expected output files to stage.

    .. data:: pre_exec

       [type: `list` | default: `[]`] Actions (shell commands) to perform before
       this task starts. Note that the set of shell commands given here are
       expected to load environments, check for work directories and data, etc.
       They are not expected to consume any significant amount of CPU time or
       other resources! Deviating from that rule will likely result in reduced
       overall throughput.

       No assumption should be made as to where these commands are executed
       (although RP attempts to perform them in the task's execution
       environment).

       No assumption should be made on the specific shell environment the
       commands are executed in.

       Errors in executing these commands will result in the task to enter
       `FAILED` state, and no execution of the actual workload will be
       attempted.

    .. data:: pre_launch

       [type: `list` | default: `[]`] Like `pre_exec`, but runs befor launching,
       potentially on a batch node which is different from the node the task is
       placed on.

    .. data:: pre_rank

       [type: `dict` | default: `{}`] A dictionary which maps a set of
       `pre_exec` like commands to each rank.  The commands are executed on the
       respective nodes where the ranks are places, and the actual rank startup
       will be delayed until all `pre_rank` commands have completed.

    .. data:: post_exec

       [type: `list` | default: `[]`] Actions (shell commands) to perform after
       this task finishes. The same remarks as on `pre_exec` apply, inclusive
       the point on error handling, which again will cause the task to fail,
       even if the actual execution was successful.

    .. data:: post_launch

       ...

    .. data:: post_rank

       ...

    .. data:: restartable

       [type: `bool` | default: `False`] If the task starts to execute on
       a pilot, but cannot finish because the pilot fails or is canceled,
       the task can be restarted.

    .. data:: scheduler

       Request the task to be handled by a specific agent scheduler


    .. data:: tags

       [type: `dict` | default: `{}`] Configuration specific tags, which
       influence task scheduling and execution (e.g., tasks co-location).

    .. data:: metadata

       [type: `ANY` | default: `None`] User defined metadata.

    .. data:: cleanup

       [type: `bool` | default: `False`] If cleanup flag is set, the pilot will
       delete the entire task sandbox upon termination. This includes all
       generated output data in that sandbox. Output staging will be performed
       before cleanup. Note that task sandboxes are also deleted if the pilot's
       own `cleanup` flag is set.

    .. data:: pilot

       [type: `str` | default: `""`] Pilot `uid`, if specified, the task is
       submitted to the pilot with the given ID. If that pilot is not known to
       the TaskManager, an exception is raised.


    Staging Directives
    ==================

    The Staging Directives are specified using a dict in the following form::

          staging_directive = {
             'source'  : None, # see 'Location' below
             'target'  : None, # see 'Location' below
             'action'  : None, # See 'Action operators' below
             'flags'   : None, # See 'Flags' below
             'priority': 0     # Control ordering of actions (unused)
          }


    Locations
    ---------

      `source` and `target` locations can be given as strings or `ru.Url`
      instances.  Strings containing `://` are converted into URLs immediately.
      Otherwise they are considered absolute or relative paths and are then
      interpreted in the context of the client's working directory.

      Special URL schemas:

        * `client://`   : relative to the client's working directory
        * `resource://` : relative to the RP    sandbox on the target resource
        * `pilot://`    : relative to the pilot sandbox on the target resource
        * `task://`     : relative to the task  sandbox on the target resource

      In all these cases, the `hostname` element of the URL is expected to be
      empty, and the path is *always* considered relative to the locations
      specified above (even though URLs usually don't have a notion of relative
      paths).


    Action operators
    ----------------

      Action operators:

        * rp.TRANSFER : remote file transfer from `source` URL to `target` URL
        * rp.COPY     : local file copy, i.e., not crossing host boundaries
        * rp.MOVE     : local file move
        * rp.LINK     : local file symlink


    Flags
    -----

      Flags:

        * rp.CREATE_PARENTS : create the directory hierarchy for targets on
          the fly
        * rp.RECURSIVE      : if `source` is a directory, handles it recursively


    Task Environment
    ================

    RP tasks are expected to be executed in isolation, meaning that their
    runtime environment is completely independent from the environment of other
    tasks, independent from the launch mechanism used to start the task, and
    also independent from the environment of the RP stack itself.

    The task description provides several hooks to help setting up the
    environment in that context.  It is important to understand the way those
    hooks interact with respect to the environments mentioned above.

      - `pre_launch` directives are set and executed before the task is passed
        on to the task launch method.  As such, `pre_launch` usually executed
        on the node where RP's agent is running, and *not* on the tasks target
        node.  Executing `pre_launch` directives for many tasks can thus
        negatively impact RP's performance (*).  Note also that `pre_launch`
        directives can in some cases interfere with the launch method.

        Use `pre_launch` directives for rare, heavy-weight operations which
        prepare the runtime environment for multiple tasks: fetch data from
        a remote source, unpack input data, create global communication
        channels, etc.

      - `pre_exec` directives are set and executed *after* the launch method
        placed the task on the compute nodes and are thus running on the target
        node.  Note that for MPI tasks, the `pre_exec` directives are executed
        once per rank.  Running large numbers of `pre_exec` directives
        concurrently can lead to system performance degradation (*), for example
        when those  directives concurrently hot the shared files system (for
        loading modules or Python virtualenvs etc).

        Use `pre_exec` directives for task environment setup such as `module
        load`, `virtualenv activate`, `export` whose effects are expected to be
        applied to all task ranks.  Avoid file staging operations at this point
        (files would be redundantly staged multiple times - once per rank).

      - `pre_rank` directives are executed only for the specified rank.  Note
        that environment settings specified in `pre_rank` will thus only apply
        to that specific rank, not to other ranks.  All other ranks stall until
        the last `pre_rank` directive has been completed -- only then is the
        actual workload task being executed on all ranks.

        Use `pre_rank` directives on rank 0 to prepare task input data: the
        operations are performed once per task, and all ranks will stall until
        the directives are completed, i.e., until input data are available.
        Avoid using `pre_rank` directives for environment settings - `pre_exec`
        will generally perform better in this case.

    (*) The performance impact of repeated concurrent access to the system's
    shared file system can be significant and can pose a major bottleneck for
    your application.  Specifically `module load` and `virtualenv activate`
    operations and the like are heavy on file system I/O, and executing those
    for many tasks is ill advised.  Having said that: RP attempts to optimize
    those operations: if it identifies that identical `pre_exec` directives are
    shared between multiple tasks, RP will execute the directives exactly *once*
    and will cache the resulting environment settings - those cached settings
    are then applied to all other tasks with the same directives, without
    executing the directives again.
    """

    _schema = {
        UID             : str         ,
        NAME            : str         ,
        MODE            : str         ,

        EXECUTABLE      : str         ,
        ARGUMENTS       : [str]       ,
        CODE            : str         ,
        FUNCTION        : str         ,
        ARGS            : [None]      ,
        KWARGS          : {str: None} ,
        COMMAND         : str         ,

        SANDBOX         : str         ,
        ENVIRONMENT     : {str: str}  ,
        NAMED_ENV       : str         ,
        PRE_LAUNCH      : [str]       ,
        PRE_EXEC        : [str]       ,
        PRE_RANK        : {int: None} ,
        POST_LAUNCH     : [str]       ,
        POST_EXEC       : [str]       ,
        POST_RANK       : {int: None} ,
        STDOUT          : str         ,
        STDERR          : str         ,
        INPUT_STAGING   : [None]      ,
        OUTPUT_STAGING  : [None]      ,
        STAGE_ON_ERROR  : bool        ,

        RANKS           : int         ,
        CORES_PER_RANK  : int         ,
        GPUS_PER_RANK   : int         ,
        THREADING_TYPE  : str         ,
        GPU_TYPE        : str         ,
        LFS_PER_RANK    : int         ,
        MEM_PER_RANK    : int         ,

        # deprecated
        CPU_PROCESSES   : int         ,  # RANKS
        CPU_PROCESS_TYPE: str         ,  # n/a
        CPU_THREADS     : int         ,  # CORES_PER_RANK
        CPU_THREAD_TYPE : str         ,  # THREADING_TYPE
        GPU_PROCESSES   : int         ,  # GPUS_PER_RANK
        GPU_PROCESS_TYPE: str         ,  # GPU_TYPE
        GPU_THREADS     : int         ,  # n/a
        GPU_THREAD_TYPE : str         ,  # n/a
        LFS_PER_PROCESS : int         ,  # LFS_PER_RANK
        MEM_PER_PROCESS : int         ,  # MEM_PER_RANK

        RESTARTABLE     : bool        ,
        SCHEDULER       : str         ,
        TAGS            : {None: None},
        METADATA        : None        ,
        CLEANUP         : bool        ,
        PILOT           : str         ,
    }

    _defaults = {
        UID             : ''          ,
        NAME            : ''          ,
        MODE            : None        ,
        EXECUTABLE      : ''          ,
        ARGUMENTS       : list()      ,
        CODE            : ''          ,
        FUNCTION        : ''          ,
        ARGS            : list()      ,
        KWARGS          : dict()      ,
        COMMAND         : ''          ,

        SANDBOX         : ''          ,
        ENVIRONMENT     : dict()      ,
        NAMED_ENV       : ''          ,
        PRE_LAUNCH      : list()      ,
        PRE_EXEC        : list()      ,
        PRE_RANK        : dict()      ,
        POST_LAUNCH     : list()      ,
        POST_EXEC       : list()      ,
        POST_RANK       : dict()      ,
        STDOUT          : ''          ,
        STDERR          : ''          ,
        INPUT_STAGING   : list()      ,
        OUTPUT_STAGING  : list()      ,
        STAGE_ON_ERROR  : False       ,

        RANKS           : 1           ,
        CORES_PER_RANK  : 1           ,
        GPUS_PER_RANK   : 0           ,
        THREADING_TYPE  : ''          ,
        GPU_TYPE        : ''          ,
        LFS_PER_RANK    : 0           ,
        MEM_PER_RANK    : 0           ,

        # deprecated
        CPU_PROCESSES   : 0           ,
        CPU_PROCESS_TYPE: ''          ,
        CPU_THREADS     : 0           ,
        CPU_THREAD_TYPE : ''          ,
        GPU_PROCESSES   : 0           ,
        GPU_PROCESS_TYPE: ''          ,
        GPU_THREADS     : 0           ,
        GPU_THREAD_TYPE : ''          ,
        LFS_PER_PROCESS : 0           ,
        MEM_PER_PROCESS : 0           ,

        RESTARTABLE     : False       ,
        SCHEDULER       : ''          ,
        TAGS            : dict()      ,
        METADATA        : None        ,
        CLEANUP         : False       ,
        PILOT           : ''          ,
    }


    # --------------------------------------------------------------------------
    #
    def __init__(self, from_dict=None):

        super().__init__(from_dict=from_dict)


    # --------------------------------------------------------------------------
    #
    def _verify(self):

        if not self.get('mode'):
            self['mode'] = TASK_EXECUTABLE

        if self.mode in [TASK_EXECUTABLE, RAPTOR_MASTER, RAPTOR_WORKER]:
            if not self.get('executable'):
                raise ValueError("TASK_EXECUTABLE Task needs 'executable'")

        elif self.mode == TASK_FUNCTION:
            if not self.get('function'):
                raise ValueError("TASK_FUNCTION Task needs 'function'")

        elif self.mode == TASK_PROC:
            if not self.get('executable'):
                raise ValueError("TASK_PROC Task needs 'executable'")

        elif self.mode == TASK_EVAL:
            if not self.get('code'):
                raise ValueError("TASK_EVAL Task needs 'code'")

        elif self.mode == TASK_EXEC:
            if not self.get('code'):
                raise ValueError("TASK_EXEC Task needs 'code'")

        elif self.mode == TASK_SHELL:
            if not self.get('command'):
                raise ValueError("TASK_SHELL Task needs 'command'")

        # backward compatibility for deprecated attributes
        if self.cpu_processes:
            self.ranks = self.cpu_processes
            self.cpu_processes = 0

        if self.cpu_threads:
            self.cores_per_rank = self.cpu_threads
            self.cpu_threads = 0

        if self.cpu_thread_type:
            self.threading_type = self.cpu_thread_type
            self.cpu_thread_type = None

        if self.gpu_processes:
            self.gpus_per_rank = self.gpu_processes
            self.gpu_processes = 0

        if self.gpu_process_type:
            self.gpu_type = self.gpu_process_type
            self.gpu_process_type = 0

        if self.lfs_per_process:
            self.lfs_per_rank = self.lfs_per_process
            self.lfs_per_process = 0

        if self.mem_per_process:
            self.mem_per_rank = self.mem_per_process
            self.mem_per_process = 0

        # deprecated and ignored
        if self.cpu_process_type: pass
        if self.gpu_threads     : pass
        if self.gpu_thread_type : pass

      # if self.mode in [TASK_SHELL, TASK_PROC]:
      #
      #     if self.get('cpu_processes', 1) * self.get('cpu_threads', 1) > 1:
      #         raise ValueError("TASK_SHELL and TASK_PROC Tasks must be single core")
      #
      #     if self.get('gpu_processes', 0) > 0:
      #         raise ValueError("TASK_SHELL and TASK_PROC Tasks canont use GPUs")


# ------------------------------------------------------------------------------



#!/usr/bin/env python3

__author__    = 'RADICAL-Cybertools Team'
__email__     = 'info@radical-cybertools.org'
__copyright__ = 'Copyright 2013-22, The RADICAL-Cybertools Team'
__license__   = 'MIT'


''' Setup script, only usable via pip. '''

import re
import os
import sys
import glob
import shutil

import subprocess as sp

from setuptools import setup, Command, find_namespace_packages


# ------------------------------------------------------------------------------
name     = 'radical.pilot'
mod_root = 'src/radical/pilot/'

# ------------------------------------------------------------------------------
#
# pip warning:
# "In-tree builds are now default. pip 22.1 will enforce this behaviour change.
#  A possible replacement is to remove the --use-feature=in-tree-build flag."
#
# With this change we need to make sure to clean out all temporary files from
# the src tree. Specifically create (and thus need to clean)
#   - VERSION
#   - SDIST
#   - the sdist file itself (a tarball)
#
# `pip install` (or any other direct or indirect invocation of `setup.py`) will
# in fact run `setup.py` multiple times: one on the top level, and internally
# again with other arguments to build sdist and bwheel packages.  We must *not*
# clean out temporary files in those internal runs as that would invalidate the
# install.
#
# We thus introduce an env variable `SDIST_LEVEL` which allows us to separate
# internal calls from the top level invocation - we only clean on the latter
# (see end of this file).
sdist_level = int(os.environ.get('SDIST_LEVEL', 0))
os.environ['SDIST_LEVEL'] = str(sdist_level + 1)


# ------------------------------------------------------------------------------
#
def sh_callout(cmd):

    p = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)

    stdout, stderr = p.communicate()
    ret            = p.returncode
    return stdout, stderr, ret


# ------------------------------------------------------------------------------
#
# versioning mechanism:
#
#   - version:          1.2.3            - is used for installation
#   - version_detail:  v1.2.3-9-g0684b06 - is used for debugging
#   - version is read from VERSION file in src_root, which then is copied to
#     module dir, and is getting installed from there.
#   - version_detail is derived from the git tag, and only available when
#     installed from git.  That is stored in mod_root/VERSION in the install
#     tree.
#   - The VERSION file is used to provide the runtime version information.
#
def get_version(_mod_root):
    '''
    a VERSION file containes the version strings is created in mod_root,
    during installation.  That file is used at runtime to get the version
    information.
    '''

    try:

        _version_base   = None
        _version_detail = None
        _sdist_name     = None

        # get version from './VERSION'
        src_root = os.path.dirname(__file__)
        if not src_root:
            src_root = '.'

        with open(src_root + '/VERSION', 'r', encoding='utf-8') as f:
            _version_base = f.readline().strip()

        # attempt to get version detail information from git
        # We only do that though if we are in a repo root dir,
        # ie. if 'git rev-parse --show-prefix' returns an empty string --
        # otherwise we get confused if the ve lives beneath another repository,
        # and the pip version used uses an install tmp dir in the ve space
        # instead of /tmp (which seems to happen with some pip/setuptools
        # versions).
        out, _, ret = sh_callout(
            'cd %s ; '
            'test -z `git rev-parse --show-prefix` || exit -1; '
            'tag=`git describe --tags --always` 2>/dev/null ; '
            'branch=`git branch | grep -e "^*" | cut -f 2- -d " "` 2>/dev/null ; '
            'echo $tag@$branch' % src_root)
        _version_detail = out.strip()
        _version_detail = _version_detail.decode()
        _version_detail = _version_detail.replace('detached from ', 'detached-')

        # remove all non-alphanumeric (and then some) chars
        _version_detail = re.sub('[/ ]+', '-', _version_detail)
        _version_detail = re.sub('[^a-zA-Z0-9_+@.-]+', '', _version_detail)

        if ret              !=  0  or \
            _version_detail == '@' or \
            'git-error'      in _version_detail or \
            'not-a-git-repo' in _version_detail or \
            'not-found'      in _version_detail or \
            'fatal'          in _version_detail :
            _version = _version_base
        elif '@' not in _version_base:
            _version = '%s-%s' % (_version_base, _version_detail)
        else:
            _version = _version_base

        # make sure the version files exist for the runtime version inspection
        _path = '%s/%s' % (src_root, _mod_root)
        with open(_path + '/VERSION', 'w', encoding='utf-8') as f:
            f.write(_version_base + '\n')
            f.write(_version      + '\n')

        _sdist_name = '%s-%s.tar.gz' % (name, _version_base)
      # _sdist_name = _sdist_name.replace('/', '-')
      # _sdist_name = _sdist_name.replace('@', '-')
      # _sdist_name = _sdist_name.replace('#', '-')
      # _sdist_name = _sdist_name.replace('_', '-')

        if '--record'    in sys.argv or \
           'bdist_egg'   in sys.argv or \
           'bdist_wheel' in sys.argv    :
            # pip install stage 2 or easy_install stage 1
            #
            # pip install will untar the sdist in a tmp tree.  In that tmp
            # tree, we won't be able to derive git version tags -- so we pack
            # the formerly derived version as ./VERSION
            shutil.move('VERSION', 'VERSION.bak')              # backup
            shutil.copy('%s/VERSION' % _path, 'VERSION')       # version to use
            os.system  ('python3 setup.py sdist')              # build sdist
            shutil.copy('dist/%s' % _sdist_name,
                        '%s/%s'   % (_mod_root, _sdist_name))  # copy into tree
            shutil.move('VERSION.bak', 'VERSION')              # restore version

        with open(_path + '/SDIST', 'w', encoding='utf-8') as f:
            f.write(_sdist_name + '\n')

        return _version_base, _version_detail, _sdist_name, _path

    except Exception as e:
        raise RuntimeError('Could not extract/set version: %s' % e) from e


# ------------------------------------------------------------------------------
# get version info -- this will create VERSION and srcroot/VERSION
version, version_detail, sdist_name, path = get_version(mod_root)


# ------------------------------------------------------------------------------
# check python version, should be >= 3.6
if sys.hexversion < 0x03060000:
    raise RuntimeError('ERROR: %s requires Python 3.6 or newer' % name)


# ------------------------------------------------------------------------------
#
class RunTwine(Command):
    user_options = []
    def initialize_options(self): pass
    def finalize_options(self):   pass
    def run(self):
        _, _, ret = sh_callout('python3 setup.py sdist upload -r pypi')
        raise SystemExit(ret)


# ------------------------------------------------------------------------------
#
# This copies the contents like examples/ dir under sys.prefix/share/$name
# It needs the MANIFEST.in entries to work.
base = 'share/%s' % name
df = [('%s/'                      % base, ['docs/source/events.md']),
      ('%s/examples'              % base, glob.glob('examples/[01]*.py')),
      ('%s/examples'              % base, glob.glob('examples/hello*')),
      ('%s/examples'              % base, glob.glob('examples/*.json')),
      ('%s/examples/docs'         % base, glob.glob('examples/docs/*.py')),
      ('%s/examples/misc'         % base, glob.glob('examples/misc/*.py')),
      ('%s/examples/misc'         % base, glob.glob('examples/misc/*.cfg')),
      ('%s/examples/data_staging' % base, glob.glob('examples/data_staging/*.py')),
]


# ------------------------------------------------------------------------------
#
setup_args = {
    'name'               : name,
    'namespace_packages' : ['radical'],
    'version'            : version,
    'description'        : 'The RADICAL pilot job framework',
    'author'             : 'RADICAL Group at Rutgers University',
    'author_email'       : 'radical@rutgers.edu',
    'maintainer'         : 'The RADICAL Group',
    'maintainer_email'   : 'radical@rutgers.edu',
    'url'                : 'https://www.github.com/radical-cybertools/radical.pilot/',
    'license'            : 'MIT',
    'keywords'           : 'radical pilot job saga',
    'python_requires'    : '>=3.6',
    'classifiers'        : [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Utilities',
        'Topic :: System :: Distributed Computing',
        'Topic :: Scientific/Engineering',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: Unix'
    ],
    'packages'           : find_namespace_packages('src', include=['radical.*']),
    'package_dir'        : {'': 'src'},
    'scripts'            : [
                            'bin/radical-pilot-agent',
                            'bin/radical-pilot-agent-statepush',
                            'bin/radical-pilot-bridge',
                            'bin/radical-pilot-bson2json',
                            'bin/radical-pilot-cleanup',
                            'bin/radical-pilot-close-session',
                            'bin/radical-pilot-component',
                            'bin/radical-pilot-create-static-ve',
                            'bin/radical-pilot-deploy-ompi.sh',
                            'bin/radical-pilot-fetch-db',
                            'bin/radical-pilot-fetch-json',
                            'bin/radical-pilot-fetch-logfiles',
                            'bin/radical-pilot-fetch-profiles',
                            'bin/radical-pilot-hello.sh',
                            'bin/radical-pilot-inspect',
                            'bin/radical-pilot-limits.py',
                            'bin/radical-pilot-prte2prof',
                            'bin/radical-pilot-resources',
                            'bin/radical-pilot-run-session',
                            'bin/radical-pilot-stats',
                            'bin/radical-pilot-stats.plot',
                            'bin/radical-pilot-ve',
                            'bin/radical-pilot-version',
                            'bin/radical-pilot-worker',
                           ],
    'package_data'       : {'': ['*.txt', '*.sh', '*.json', '*.gz', '*.c',
                                 '*.md', 'VERSION', 'SDIST', sdist_name]},
  # 'setup_requires'     : ['pytest-runner'],
    'install_requires'   : ['radical.utils>=1.12',
                            'radical.saga>=1.12',
                            'radical.gtod',
                            'pymongo<4',
                            'setproctitle',
                            'dill'
                           ],
    'tests_require'      : ['pytest',
                            'pylint',
                            'flake8',
                            'coverage',
                            'mock==2.0.0.',
                           ],
    'test_suite'         : '%s.tests' % name,
    'zip_safe'           : False,
  # 'build_sphinx'       : {
  #     'source-dir'     : 'docs/',
  #     'build-dir'      : 'docs/build',
  #     'all_files'      : 1,
  # },
  # 'upload_sphinx'      : {
  #     'upload-dir'     : 'docs/build/html',
  # },
    # This copies the contents of the examples/ dir under
    # sys.prefix/share/$name
    # It needs the MANIFEST.in entries to work.
    'data_files'         : df,
    'cmdclass'           : {'upload': RunTwine},
}


# ------------------------------------------------------------------------------
#
setup(**setup_args)


# ------------------------------------------------------------------------------
# clean temporary files from source tree
if sdist_level == 0:
    os.system('rm -vrf src/%s.egg-info' % name)
    os.system('rm -vf  %s/%s'           % (path, sdist_name))
    os.system('rm -vf  %s/VERSION'      % path)
    os.system('rm -vf  %s/SDIST'        % path)


# ------------------------------------------------------------------------------





'''
The serializer should be able to (de)serialize information that we want
to send over the wire from the client side to the agent side via
1- ZMQ
2- MongoDB

we except:
    1- Callables with and without dependecies.
    2- Non-callables like classes and other python objects
'''
__copyright__ = 'Copyright 2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os
import dill
import pickle
import codecs
import tempfile


_obj_dir       = tempfile.gettempdir()
_obj_file_name = 'rp_obj.pkl'
_obj_file_path = os.path.join(_obj_dir, _obj_file_name)


# ------------------------------------------------------------------------------
#
def serialize_obj(obj):
    '''
    serialize object
    '''
    result    = None
    exception = None

    if callable(obj):
        try:
            result = dill.dumps(obj)
        except Exception as e:
            exception = e

        # if we fail, then pikle it by reference
        if result is None:
            # see issue: https://github.com/uqfoundation/dill/issues/128
            try:
                result = dill.dumps(obj, byref = True)
            except Exception as e:
                exception = e

    else:
        try:
            result = dill.dumps(obj, recurse = True)
        except Exception as e:
            exception = e

        if result is None:
            # see issue: https://github.com/uqfoundation/dill/issues/128
            try:
                result = dill.dumps(obj, byref = True)
            except Exception as e:
                exception = e

    if result is None:
        raise Exception("object %s is not serializable") from exception

    return  result


# ------------------------------------------------------------------------------
#
def serialize_file(obj):
    '''
    serialize object to file
    '''

    # FIXME: assign unique path and id for the pickled file
    #        to avoid overwriting to the same file

    result    = None
    exception = None

    kwargs = {}
    if not callable(obj):
        kwargs['recurse'] = True

    try:
        with open(_obj_file_path, 'wb') as f:
            dill.dump(obj, f, **kwargs)

    except Exception as e:
        exception = e

    else:
        result = _obj_file_path

    if result is None:
        raise Exception('object is not serializable') from exception

    return _obj_file_path


# ------------------------------------------------------------------------------
#
def deserialize_file(fname):
    '''
    Deserialize object from file
    '''
    result = None

    # check if we have a valid file
    if not os.path.isfile(fname):
        return None

    try:
        with open(fname, 'rb') as f:
            result = dill.load(f)
    except Exception as e:
        raise Exception ("failed to deserialize object from file") from e

    if result is None:
        raise RuntimeError('failed to deserialize')
    return result


# ------------------------------------------------------------------------------
#
def deserialize_obj(obj):
    '''
    Deserialize object from str.

    Raises
    ------
    TypeError
        if object cannot be deserialized.
    ValueError
        if no result is found.

    '''
    result = None

    try:
        result = dill.loads(obj)
    except Exception as e:
        raise TypeError("failed to deserialize from object") from e

    # Ref https://github.com/radical-cybertools/radical.pilot/pull/ \
    #                        2615#discussion_r870356482
    assert result is not None
    return result


# ------------------------------------------------------------------------------
#
def serialize_bson(obj):

    result = codecs.encode(pickle.dumps(obj), "base64").decode()

    return result


# ------------------------------------------------------------------------------
#
def deserialize_bson(obj):

    result = pickle.loads(codecs.decode(obj.encode(), "base64"))

    return result


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import os
import sys

import threading

import radical.utils as ru


# ------------------------------------------------------------------------------
#
def log(msg):

    sys.stdout.write('%s\n' % msg)
    sys.stdout.flush()


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    uid       = os.environ['RP_TASK_ID']
    addr_work = os.environ['RP_WORK_QUEUE_IN']
    addr_res  = os.environ['RP_RESULT_QUEUE_OUT']

    n_clients = int(sys.argv[1])
    term      = threading.Event()
    term_cnt  = 0

    work_queue   = ru.Putter('RP_WORK_QUEUE',   addr_work)
    result_queue = ru.Getter('RP_RESULT_QUEUE', addr_res)

    for i in range(1000):
        log('> %06d ?' % (i))
        work_queue.put(i)

    for i in range(1000):
        result = result_queue.get()
        log('< %06d : %s' % (i, result))

    for i in range(n_clients):
        log('x %02d QUIT' % i)
        work_queue.put('QUIT')

    log('m DONE %s')
    sys.exit()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import sys
import radical.pilot as rp

# READ: The RADICAL-Pilot documentation:
#   https://radicalpilot.readthedocs.io/en/stable/
#
# Try running this example with RADICAL_PILOT_VERBOSE=debug set if
# you want to see what happens behind the scenes!


# ------------------------------------------------------------------------------
#
def pilot_state_cb (pilot, state):
    """ this callback is invoked on all pilot state changes """

    print("[Callback]: Pilot '%s' state: %s." % (pilot.uid, state))

    if state == rp.FAILED:
        sys.exit (1)


# ------------------------------------------------------------------------------
#
def task_state_cb (task, state):
    """ this callback is invoked on all task state changes """

    print("[Callback]: Task '%s' state: %s." % (task.uid, state))

    if state == rp.FAILED:
        sys.exit (1)


# ------------------------------------------------------------------------------
#
if __name__ == "__main__":

    # we can optionally pass session name to RP
    if len(sys.argv) > 1:
        session_name = sys.argv[1]
    else:
        session_name = None

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session(uid=session_name)
    print("session id: %s" % session.uid)

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        # Add a Pilot Manager. Pilot managers manage one or more Pilots.
        pmgr = rp.PilotManager(session=session)

        # Register our callback with the PilotManager. This callback will get
        # called every time any of the pilots managed by the PilotManager
        # change their state.
        pmgr.register_callback(pilot_state_cb)

        # Define a single-core local pilot that runs for 5 minutes and cleans up
        # after itself.
        pdesc = rp.PilotDescription()
        pdesc.resource = "local.localhost"
        pdesc.cores    = 1
        pdesc.runtime  = 5

        # Launch the pilot.
        pilot = pmgr.submit_pilots(pdesc)

        # Create a Taskt that sorts the local password file and writes the
        # output to result.dat.
        #
        #  The exact command that is executed by the agent is:
        #    "/usr/bin/sort -o result.dat passwd"
        #
        td = rp.TaskDescription()
        td.executable     = "/usr/bin/sort"
        td.arguments      = ["-o", "result.dat", "passwd"]
        td.input_staging  = "/etc/passwd"
        td.output_staging = "result.dat"

        # Combine the Pilot, the Tasks and a scheduler via
        # a TaskManager object.
        tmgr = rp.TaskManager(session=session)

        # Register our callback with the TaskManager. This callback will get
        # called every time any of the tasks managed by the TaskManager
        # change their state.
        tmgr.register_callback(task_state_cb)

        # Add the previously created Pilot to the TaskManager.
        tmgr.add_pilots(pilot)

        # Submit the previously created Task description to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning the Task to the Pilot.
        task = tmgr.submit_tasks(td)

        # Wait for the task to reach a terminal state (DONE or FAILED).
        tmgr.wait_tasks()

        print("* Task %s state: %s, exit code: %s,"
              % (task.uid, task.state, task.exit_code))

    except Exception as e:
        # Something unexpected happened in the pilot code above
        print("caught Exception: %s" % e)
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        print("need to exit now: %s" % e)

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.
        print("closing session")
        session.close ()

        # the above is equivalent to
        #
        #   session.close (cleanup=True, terminate=True)
        #
        # it will thus both clean out the session's database record, and kill
        # all remaining pilots (none in our example).


# -------------------------------------------------------------------------------





__copyright__ = 'Copyright 2018-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

from .base import RMInfo, ResourceManager


# ------------------------------------------------------------------------------
#
class Debug(ResourceManager):

    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, rm_info: RMInfo) -> RMInfo:

        nodes = [('localhost', rm_info.cores_per_node)
                 for idx in range(rm_info.requested_nodes)]

        rm_info.node_list = self._get_node_list(nodes, rm_info)

        # UIDs need to be made unique
        for idx, node in enumerate(rm_info.node_list):
            node['node_id'] = '%s_%04d' % (node['node_name'], idx)

        return rm_info


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2014, http://radical.rutgers.edu'
__license__   = 'MIT'

import os
import sys

import radical.pilot as rp
import radical.utils as ru


pwd = os.path.dirname(os.path.abspath(__file__))
dh  = ru.DebugHelper()


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    # die - this is where the restart attempt will continue
    # but before, we store some UIDs
    with ru.ru_open('%s/restart.dat' % pwd, 'r') as fin:
        session_id = fin.readline().split()[1].strip()
        tmgr_id    = fin.readline().split()[1].strip()
        pmgr_id    = fin.readline().split()[1].strip()
        pilot_id   = fin.readline().split()[1].strip()
        task_ids   = fin.readline().split()[1].strip().split(':')

    print('session_id : %s' % session_id)
    print('tmgr_id    : %s' % tmgr_id)
    print('pmgr_id    : %s' % pmgr_id)
    print('pilot_id   : %s' % pilot_id)
    print('task_ids   : %s' % task_ids)

    # we use a reporter class for nicer output
    report = ru.Reporter(name='radical.pilot')
    report.title('Getting Started (RP version %s)' % rp.version)

    # use the resource specified as argument, fall back to localhost
    if   len(sys.argv)  > 2: report.exit('Usage:\t%s [resource]\n\n' % sys.argv[0])
    elif len(sys.argv) == 2: resource = sys.argv[1]
    else                   : resource = 'local.localhost'

    try:
        # Create a new session, and reconnect tmgr and pmgr
        session = rp.Session(uid=session_id)
        tmgr    = rp.TaskManager(session=session,  uid=tmgr_id)
        pmgr    = rp.PilotManager(session=session, uid=pmgr_id)

        # re-add pilots to the uymgr
        tmgr.add_pilots(pmgr.get_pilots())

        print(pmgr.list_pilots())
        print(tmgr.list_tasks())

        for task in tmgr.get_tasks():
            print(task.uid, task.state)


        # Wait for all compute tasks to reach a final state (DONE, CANCELED or FAILED).
        report.header('gather results')
        tmgr.wait_tasks()


    except Exception as e:
        # Something unexpected happened in the pilot code above
        report.error('caught Exception: %s\n' % e)
        ru.print_exception_trace()
        raise

    except (KeyboardInterrupt, SystemExit):
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        ru.print_exception_trace()
        report.warn('exit requested\n')

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.  This will kill all remaining pilots.
        report.header('finalize')
        session.close(download=True)

    report.header()


# ------------------------------------------------------------------------------

